# coding: utf-8

from .event import StreamEvent
from .stream_connection import StreamConnection, LOGIN_STREAM_ID
from .stream_cxn_state import StreamCxnState


class OMMStreamConnection(StreamConnection):
    @property
    def subprotocol(self) -> str:
        return "tr_json2"

    def get_login_message(self):
        login_message = {
            "Domain": "Login",
            "ID": LOGIN_STREAM_ID,
            "Key": self._session.get_omm_login_message(),
        }
        return login_message

    def get_close_message(self) -> dict:
        close_message = {
            "Domain": "Login",
            "ID": LOGIN_STREAM_ID,
            "Type": "Close",
        }
        return close_message

    def _handle_login_message(self, message: dict):
        """
        Parameters
        ----------
        message
            {
                'ID': 2,
                'Type': 'Refresh',
                'Domain': 'Login',
                'Key':
                    {
                        'Name': TOKEN_HERE,
                        'Elements': {
                            'AllowSuspectData': 1, 'ApplicationId': '256',
                            'ApplicationName': 'RTO',
                            'AuthenticationErrorCode': 0,
                            'AuthenticationErrorText': {
                                'Type': 'AsciiString', 'Data': None
                            },
                            'AuthenticationTTReissue': 1634562361,
                            'Position': '10.46.188.21/EPUAKYIW3629',
                            'ProvidePermissionExpressions': 1,
                            'ProvidePermissionProfile': 0,
                            'SingleOpen': 1, 'SupportEnhancedSymbolList': 1,
                            'SupportOMMPost': 1,
                            'SupportPauseResume': 0, 'SupportStandby': 0,
                            'SupportBatchRequests': 7,
                            'SupportViewRequests': 1, 'SupportOptimizedPauseResume': 0
                        }
                    },
                'State':
                    {
                        'Stream': 'Open', 'Data': 'Ok',
                        'Text': 'Login accepted by host ads-fanout-lrg-az2-apse1-prd.'
                    }, 'Elements': {'PingTimeout': 30, 'MaxMsgSize': 61426}
            }
        """

        state = message.get("State", {})
        stream_state = state.get("Stream")

        if stream_state == "Open":
            self._state = StreamCxnState.Ready
            self.prepared.set_result(True)

        elif stream_state == "Closed":
            self.error(
                f"{self._classname} received a closing message "
                f"(cxn state={self.state}, message state={state})"
            )
            self._state = StreamCxnState.Disconnected
            self.dispose()

        else:
            raise ValueError(
                f"{self._classname} Don't know what to do "
                f"(cxn state={self.state}, "
                f"stream_state={stream_state}, message={message})"
            )

    def _process_message(self, message: dict) -> None:
        self.debug(f"{self._classname} process_message {message} ")

        message_type = message.get("Type")
        stream_id = message.get("ID")

        event = StreamEvent.get(stream_id)

        if message_type == "Refresh":
            self._emitter.emit(event.refresh_by_id, self, message)
            message_complete = message.get("Complete", True)
            message_complete and self._emitter.emit(event.complete_by_id, self, message)

        elif message_type == "Update":
            self._emitter.emit(event.update_by_id, self, message)

        elif message_type == "Status":
            self._emitter.emit(event.status_by_id, self, message)

        elif message_type == "Error":
            self._emitter.emit(event.error_by_id, self, message)

        elif message_type == "Ping":
            pong_message = {"Type": "Pong"}
            self.send(pong_message)

        else:
            raise ValueError(f"Unknown message {message}")



from typing import Iterable, Union, Optional, TYPE_CHECKING

from ..._content_provider import ContentProviderLayer
from ..._content_type import ContentType
from ...._tools import validate_types, validate_bool_value

if TYPE_CHECKING:
    from ....content._types import ExtendedParams


class Definition(ContentProviderLayer):
    """
    This class describe parameters to retrieve the calculated concentration data by all consolidated investors.

    Parameters
    ----------
    universe: str, list of str
        The Universe parameter allows the user to define the companies for which the content is returned.

    limit: int, optional
        The limit parameter is used for paging. It allows users to select the number of records to be returned.
        Default page size is 100 or 20 (depending on the operation).

    use_field_names_in_headers: bool, optional
        Return field name as column headers for data instead of title

    extended_params : ExtendedParams, optional
        If necessary other parameters.

    Examples
    --------
    >>> from refinitiv.data.content import ownership
    >>> definition = ownership.consolidated.investors.Definition("TRI.N")
    >>> response = definition.get_data()
    """

    def __init__(
        self,
        universe: Union[str, Iterable[str]],
        limit: Optional[int] = None,
        use_field_names_in_headers: bool = False,
        extended_params: "ExtendedParams" = None,
    ):
        validate_types(limit, [int, type(None)], "limit")
        validate_bool_value(use_field_names_in_headers)

        super().__init__(
            content_type=ContentType.OWNERSHIP_CONSOLIDATED_INVESTORS,
            universe=universe,
            limit=limit,
            use_field_names_in_headers=use_field_names_in_headers,
            extended_params=extended_params,
        )



from typing import Any, Callable, TYPE_CHECKING

from ._quantitative_data_stream import QuantitativeDataStream
from ..._types import ExtendedParams
from ...._core.session import get_valid_session
from ...._tools import cached_property, create_repr, make_callback
from ....delivery._stream.base_stream import StreamOpenMixin

if TYPE_CHECKING:
    from pandas import DataFrame
    from ...._core.session import Session


class Stream(StreamOpenMixin):
    """
    Open a streaming quantitative analytic service subscription.

    Parameters
    ----------
    universe: dict

    fields: list

    extended_params: dict
        Default: None

    Methods
    -------
    open()
        Open the QuantitativeDataStream connection

    close()
        Close the QuantitativeDataStream connection, releases resources

    get_snapshot()
        Get DataFrame with stream

    """

    def __init__(
        self,
        universe: dict,
        fields: list = None,
        session: "Session" = None,
        extended_params: ExtendedParams = None,
    ) -> None:
        session = get_valid_session(session)
        self._universe = universe
        self._fields = fields
        self._session = session
        self._extended_params = extended_params

    @cached_property
    def _stream(self) -> QuantitativeDataStream:
        stream = QuantitativeDataStream(
            universe=self._universe,
            fields=self._fields,
            session=self._session,
            extended_params=self._extended_params,
        )
        return stream

    def get_snapshot(self) -> "DataFrame":
        """
        Returns DataFrame snapshot a streaming quantitative analytic service

        Returns
        -------
        pd.DataFrame
        """
        return self._stream.get_snapshot()

    def on_response(self, func: Callable[[list, list, "Stream"], Any]) -> "Stream":
        """
        This callback is called with the reference to the stream object,
        the instrument name and the instrument response

        Parameters
        ----------
        func : Callable
            Called when the stream has response

        Returns
        -------
        current instance
        """
        self._stream.on_response(make_fin_callback(func))
        return self

    def on_update(self, func: Callable[[list, list, "Stream"], Any]) -> "Stream":
        """
        This callback is called with the reference to the stream object,
        the instrument name and the instrument update

        Parameters
        ----------
        func : Callable
            Called when the stream has a new update

        Returns
        -------
        current instance
        """
        self._stream.on_update(make_fin_callback(func))
        return self

    def on_state(self, func: Callable[[list, "Stream"], Any]) -> "Stream":
        """
        This callback is called with the reference to the stream object,
        when the stream has new state

        Parameters
        ----------
        func : Callable
            Called when the stream has a new state

        Returns
        -------
        current instance
        """
        self._stream.on_alarm(make_callback(func))
        return self

    def __repr__(self):
        return create_repr(
            self,
            class_name=self.__class__.__name__,
        )


def make_fin_callback(func: Callable[["Stream", list, list], Any]) -> Callable:
    """Return a callback functions with correct arguments order."""

    def callback(stream, *args):
        func(*args, stream)

    return callback



# coding: utf8

__all__ = ["AverageType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class AverageType(Enum):
    ARITHMETIC_RATE = "ArithmeticRate"
    ARITHMETIC_STRIKE = "ArithmeticStrike"
    GEOMETRIC_RATE = "GeometricRate"
    GEOMETRIC_STRIKE = "GeometricStrike"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(AverageType, _AVERAGE_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_AVERAGE_TYPE_VALUES_IN_LOWER_BY_AVERAGE_TYPE, some)


_AVERAGE_TYPE_VALUES = tuple(t.value for t in AverageType)
_AVERAGE_TYPE_VALUES_IN_LOWER_BY_AVERAGE_TYPE = {
    name.lower(): item for name, item in AverageType.__members__.items()
}



# coding: utf8

from .._object_definition import ObjectDefinition
from ._enums import RiskType
from ._enums import AssetClass


class InterestRateCurveGetDefinition(ObjectDefinition):
    def __init__(
        self,
        index_name=None,
        main_constituent_asset_class=None,
        risk_type=None,
        currency=None,
        curve_tag=None,
        id=None,
        name=None,
        source=None,
        valuation_date=None,
    ):
        super().__init__()
        self.index_name = index_name
        self.main_constituent_asset_class = main_constituent_asset_class
        self.risk_type = risk_type
        self.currency = currency
        self.curve_tag = curve_tag
        self.id = id
        self.name = name
        self.source = source
        self.valuation_date = valuation_date

    @property
    def main_constituent_asset_class(self):
        """
        :return: enum AssetClass
        """
        return self._get_enum_parameter(AssetClass, "mainConstituentAssetClass")

    @main_constituent_asset_class.setter
    def main_constituent_asset_class(self, value):
        self._set_enum_parameter(AssetClass, "mainConstituentAssetClass", value)

    @property
    def risk_type(self):
        """
        :return: enum RiskType
        """
        return self._get_enum_parameter(RiskType, "riskType")

    @risk_type.setter
    def risk_type(self, value):
        self._set_enum_parameter(RiskType, "riskType", value)

    @property
    def currency(self):
        """
        The currency code of the interest rate curve
        :return: str
        """
        return self._get_parameter("currency")

    @currency.setter
    def currency(self, value):
        self._set_parameter("currency", value)

    @property
    def curve_tag(self):
        """
        User defined string to identify the curve. It can be used to link output results to the curve definition. Only alphabetic,
        numeric and '- _.#=@' characters are supported. Optional.
        :return: str
        """
        return self._get_parameter("curveTag")

    @curve_tag.setter
    def curve_tag(self, value):
        self._set_parameter("curveTag", value)

    @property
    def id(self):
        """
        Id of the curve definition
        :return: str
        """
        return self._get_parameter("id")

    @id.setter
    def id(self, value):
        self._set_parameter("id", value)

    @property
    def index_name(self):
        """
        :return: str
        """
        return self._get_parameter("indexName")

    @index_name.setter
    def index_name(self, value):
        self._set_parameter("indexName", value)

    @property
    def name(self):
        """
        The name of the interest rate curve
        :return: str
        """
        return self._get_parameter("name")

    @name.setter
    def name(self, value):
        self._set_parameter("name", value)

    @property
    def source(self):
        """
        :return: str
        """
        return self._get_parameter("source")

    @source.setter
    def source(self, value):
        self._set_parameter("source", value)

    @property
    def valuation_date(self):
        """
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)



import socket
from typing import Iterable, Optional, TYPE_CHECKING

import httpx
from appdirs import user_config_dir, user_data_dir

from .auth_manager import AuthManager
from .event_code import EventCode
from ... import __version__
from ..._tools import cached_property, urljoin
from ...errors import DesktopSessionError, PlatformSessionError

if TYPE_CHECKING:
    from ._desktop_session import DesktopSession


def update_port_in_url(url, port):
    try:
        protocol, path, default_port = url.split(":")
    except ValueError:
        protocol, path, *_ = url.split(":")

    if port is not None:
        retval = ":".join([protocol, path, str(port)])
    else:
        retval = url

    return retval


def read_firstline_in_file(filename, logger=None):
    try:
        f = open(filename)
        first_line = f.readline()
        f.close()
        return first_line
    except IOError as e:
        if logger:
            logger.error(f"I/O error({e.errno}): {e.strerror}")
        return ""


# --------------------------------------------------------------------------------------
#   Desktop
# --------------------------------------------------------------------------------------


class DesktopConnection:
    def __init__(self, session: "DesktopSession"):
        self._session = session
        self._base_url = self._session._base_url
        self._uuid = self._session._uuid
        self.app_key = self._session.app_key
        self._dp_proxy_base_url = self._session._dp_proxy_base_url

    def get_timeout(self):
        return self._session.get_timeout()

    def get_handshake_url(self):
        return self._session._get_handshake_url()

    def set_port_number(self, port_number):
        return self._session.set_port_number(port_number)

    def error(self, msg, *args, **kwargs):
        return self._session.error(msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        return self._session.debug(msg, *args, **kwargs)

    def open(self) -> bool:
        is_opened = True
        port_number = None
        try:
            if not self._dp_proxy_base_url:
                # Identify port number to update base url
                port_number = self.identify_scripting_proxy_port()
                self.set_port_number(port_number)

            handshake_url = self.get_handshake_url()

            self.handshake(handshake_url)

        except DesktopSessionError as e:
            self.error(e.message)
            is_opened = False

        if not self._dp_proxy_base_url and not port_number:
            is_opened = False
            self._session._call_on_event(
                EventCode.SessionAuthenticationFailed,
                "Eikon is not running",
            )

        return is_opened

    def close(self):
        # nothing to close
        pass

    def identify_scripting_proxy_port(self):
        """
        Returns the port used by the Scripting Proxy stored in a configuration file.
        """
        import platform
        import os

        port = None
        path = []
        func = user_config_dir if platform.system() == "Linux" else user_data_dir
        app_names = ["Data API Proxy", "Eikon API proxy", "Eikon Scripting Proxy"]
        for app_author in ["Refinitiv", "Thomson Reuters"]:
            path = path + [
                func(app_name, app_author, roaming=True)
                for app_name in app_names
                if os.path.isdir(func(app_name, app_author, roaming=True))
            ]

        if len(path):
            port_in_use_file = os.path.join(path[0], ".portInUse")

            # Test if ".portInUse" file exists
            if os.path.exists(port_in_use_file):
                # First test to read .portInUse file
                first_line = read_firstline_in_file(port_in_use_file)
                if first_line != "":
                    saved_port = first_line.strip()
                    test_proxy_url = update_port_in_url(self._base_url, saved_port)
                    test_proxy_result = self.check_proxy(test_proxy_url)
                    if test_proxy_result:
                        port = saved_port
                        self.debug(f"Port {port} was retrieved from .portInUse file")
                    else:
                        self.debug(
                            f"Retrieved port {saved_port} value "
                            f"from .portIntUse isn't valid."
                        )

        if port is None:
            self.debug(
                "Warning: file .portInUse was not found. "
                "Try to fallback to default port number."
            )
            port = self.get_port_number_from_range(("9000", "9060"), self._base_url)

        if port is None:
            self.error(
                "Error: no proxy address identified.\nCheck if Desktop is running."
            )
            return None

        return port

    def get_port_number_from_range(
        self, ports: Iterable[str], url: str
    ) -> Optional[str]:
        for port_number in ports:
            self.debug(f"Try defaulting to port {port_number}...")
            test_proxy_url = update_port_in_url(url, port_number)
            test_proxy_result = self.check_proxy(test_proxy_url)
            if test_proxy_result:
                self.debug(f"Default proxy port {port_number} was successfully checked")
                return port_number
            self.debug(f"Default proxy port #{port_number} failed")

        return None

    def check_proxy(self, url: str, timeout=None) -> bool:
        #   set default timeout
        timeout = timeout if timeout is not None else self.get_timeout()
        url = urljoin(url, "/api/status")

        try:
            response = self._session.http_request(
                url=url,
                method="GET",
                timeout=timeout,
            )

            self.debug(
                f"Checking proxy url {url} response : "
                f"{response.status_code} - {response.text}"
            )
            return True
        except (socket.timeout, httpx.ConnectTimeout):
            self.debug(f"Timeout on checking proxy url {url}")
        except ConnectionError as e:
            self.debug(f"Connexion Error on checking proxy {url} : {e!r}")
        except Exception as e:
            self.debug(f"Error on checking proxy url {url} : {e!r}")
        return False

    def handshake(self, url, timeout=None):
        #   set default timeout
        timeout = timeout if timeout is not None else self.get_timeout()
        self.debug(f"Try to handshake on url {url}...")

        try:
            # DAPI for E4 - API Proxy - Handshake
            _body = {
                "AppKey": self.app_key,
                "AppScope": "trapi",
                "ApiVersion": "1",
                "LibraryName": "RDP Python Library",
                "LibraryVersion": __version__,
            }

            if self._uuid:
                # add uuid for DP-PROXY multi user mode
                _body.update({"Uuid": self._uuid})

            _headers = {"Content-Type": "application/json"}

            response = None
            try:
                response = self._session.http_request(
                    url=url,
                    method="POST",
                    headers=_headers,
                    json=_body,
                    timeout=timeout,
                )

                self.debug(f"Response : {response.status_code} - {response.text}")
            except Exception as e:
                self.debug(f"HTTP request failed: {e!r}")

            if response:
                if response.status_code == httpx.codes.OK:
                    result = response.json()
                    self._session._access_token = result.get("access_token", None)

                elif response.status_code == httpx.codes.BAD_REQUEST:
                    self.error(
                        f"Status code {response.status_code}: "
                        f"Bad request on handshake url {url} : {response.text}"
                    )
                    key_is_incorrect_msg = (
                        f"Status code {response.status_code}: App key is incorrect"
                    )
                    self._session._call_on_event(
                        EventCode.SessionAuthenticationFailed,
                        key_is_incorrect_msg,
                    )
                    raise DesktopSessionError(1, key_is_incorrect_msg)
                else:
                    self.debug(
                        f"Response {response.status_code} on handshake url {url} : "
                        f"{response.text}"
                    )

        except (socket.timeout, httpx.ConnectTimeout):
            raise DesktopSessionError(1, f"Timeout on handshake url {url}")
        except Exception as e:
            raise DesktopSessionError(1, f"Error on handshake url {url} : {e!r}")


# --------------------------------------------------------------------------------------
#   RefinitivData
# --------------------------------------------------------------------------------------


class RefinitivDataConnection:
    def __init__(self, session):
        self._session = session

    @cached_property
    def auth_manager(self):
        return AuthManager(
            self._session,
            self._session.server_mode,
        )

    def get_omm_login_message(self) -> dict:
        return {
            "NameType": "AuthnToken",
            "Elements": {
                "AuthenticationToken": self._session._access_token,
                "ApplicationId": self._session._dacs_params.dacs_application_id,
                "Position": self._session._dacs_params.dacs_position,
            },
        }

    async def http_request_async(
        self,
        url: str,
        method=None,
        headers=None,
        data=None,
        params=None,
        json=None,
        closure=None,
        **kwargs,
    ):
        return await self._session._http_service.request_async(
            url,
            method=method,
            headers=headers,
            data=data,
            params=params,
            json=json,
            closure=closure,
            **kwargs,
        )

    def http_request(
        self,
        url: str,
        method=None,
        headers=None,
        data=None,
        params=None,
        json=None,
        closure=None,
        **kwargs,
    ):
        return self._session._http_service.request(
            url,
            method=method,
            headers=headers,
            data=data,
            params=params,
            json=json,
            closure=closure,
            **kwargs,
        )

    def open(self) -> bool:
        is_opened = self.auth_manager.authorize()
        return is_opened

    def close(self):
        self.auth_manager.close()


class RefinitivDataAndDeployedConnection(RefinitivDataConnection):
    def __init__(self, session):
        RefinitivDataConnection.__init__(self, session)

    def get_omm_login_message(self):
        return {
            "Name": self._session._dacs_params.deployed_platform_username,
            "Elements": {
                "ApplicationId": self._session._dacs_params.dacs_application_id,
                "Position": self._session._dacs_params.dacs_position,
            },
        }


class DeployedConnection(RefinitivDataAndDeployedConnection):
    def __init__(self, session):
        RefinitivDataAndDeployedConnection.__init__(self, session)

    async def http_request_async(self, *args, **kwargs):
        raise PlatformSessionError(
            -1,
            "Error!!! Platform session cannot connect to refinitiv dataplatform. "
            "Please check or provide the access right.",
        )

    def http_request(self, *args, **kwargs):
        raise PlatformSessionError(
            -1,
            "Error!!! Platform session cannot connect to refinitiv dataplatform. "
            "Please check or provide the access right.",
        )

    def open(self) -> bool:
        return True

    def close(self):
        # nothing to close
        pass



from typing import Iterable, Union, TYPE_CHECKING, Optional

from ..._types import Strings

if TYPE_CHECKING:
    from ._models import Step, Turn, Constituents
    from ._zc_curve_parameters import ZcCurveParameters
    from ._zc_curve_definitions import ZcCurveDefinitions
    from ._enums import ZcCurvesOutputs

Steps = Union[Iterable["Step"]]
Turns = Union[Iterable["Turn"]]
OptConstituents = Optional["Constituents"]
CurveDefinition = Optional["ZcCurveDefinitions"]
CurveParameters = Optional["ZcCurveParameters"]
Universe = Union["zc_curve.Definition", Iterable["zc_curve.Definition"]]
Outputs = Union[Strings, Iterable["ZcCurvesOutputs"]]



# coding: utf8

from typing import Optional

from .._base import UnderlyingDefinition
from ..._instrument_definition import InstrumentDefinition
from .._enums import (
    BuySell,
    CallPut,
    ExerciseStyle,
    UnderlyingType,
)
from ._eti_barrier_definition import EtiBarrierDefinition
from ._eti_binary_definition import EtiBinaryDefinition
from ._eti_cbbc_definition import EtiCbbcDefinition
from ._eti_double_barriers_definition import EtiDoubleBarriersDefinition
from ._eti_fixing_info import EtiFixingInfo
from ._eti_underlying_definition import EtiUnderlyingDefinition


class EtiDefinition(InstrumentDefinition):
    """
    Parameters
    ----------
    instrument_tag : str, optional
        User defined string to identify the instrument.It can be used to link output
        results to the instrument definition. Only alphabetic, numeric and '- _.#=@'
        characters are supported. Optional.
    instrument_code : str, optional
        An option RIC that is used to retrieve the description of the
        EtiOptionDefinition contract. Optional.If null, the instrument_code of
        underlying_definition must be provided.
    end_date : str, optional
        Expiry date of the option
    asian_definition : EtiOptionFixingInfo, optional
        Fixing details for asian options
    barrier_definition : EtiOptionBarrierDefinition, optional
        Details for barrier option.
    binary_definition : EtiOptionBinaryDefinition, optional
        Details for binary option.
    buy_sell : BuySell, optional
        The side of the deal. Possible values:
        - Buy
        - Sell
    call_put : CallPut, optional
        Tells if the option is a call or a put. Possible values:
        - Call
        - Put
    cbbc_definition : EtiOptionCbbcDefinition, optional
        Details for CBBC (Call Bear/Bull Contract) option.
    double_barriers_definition : EtiOptionDoubleBarriersDefinition, optional
        Details for double barriers option.
    exercise_style : ExerciseStyle, optional
        EURO or AMER
    underlying_definition : EtiUnderlyingDefinition, optional
        Details of the underlying. Can be used to override some data of the underlying.
    underlying_type : UnderlyingType, optional
        Underlying type of the option. Possible values:
        - Eti
        - Fx
    deal_contract : int, optional
        deal_contract. It is the number of contracts bought or sold in the deal.
    end_date_time : str, optional
        Expiry date time of the option
    lot_size : float, optional
        The lot size. It is the number of options bought or sold in one transaction.
    offset : int, optional
        offset. The offset in minutes between the time UTC and the time of the exchange
        where the contract is traded.
    strike : float, optional
        strike of the option
    """

    def __init__(
        self,
        instrument_tag: Optional[str] = None,
        instrument_code: Optional[str] = None,
        end_date: Optional[str] = None,
        asian_definition: Optional[EtiFixingInfo] = None,
        barrier_definition: Optional[EtiBarrierDefinition] = None,
        binary_definition: Optional[EtiBinaryDefinition] = None,
        buy_sell: Optional[BuySell] = None,
        call_put: Optional[CallPut] = None,
        cbbc_definition: Optional[EtiCbbcDefinition] = None,
        double_barriers_definition: Optional[EtiDoubleBarriersDefinition] = None,
        exercise_style: Optional[ExerciseStyle] = None,
        underlying_definition: Optional[EtiUnderlyingDefinition] = None,
        underlying_type: Optional[UnderlyingType] = None,
        deal_contract: Optional[int] = None,
        end_date_time: Optional[str] = None,
        lot_size: Optional[float] = None,
        offset: Optional[int] = None,
        strike: Optional[float] = None,
        **kwargs,
    ) -> None:
        super().__init__(instrument_tag, **kwargs)
        self.instrument_code = instrument_code
        self.end_date = end_date
        self.asian_definition = asian_definition
        self.barrier_definition = barrier_definition
        self.binary_definition = binary_definition
        self.buy_sell = buy_sell
        self.call_put = call_put
        self.cbbc_definition = cbbc_definition
        self.double_barriers_definition = double_barriers_definition
        self.exercise_style = exercise_style
        self.underlying_definition = underlying_definition
        self.underlying_type = underlying_type
        self.deal_contract = deal_contract
        self.end_date_time = end_date_time
        self.lot_size = lot_size
        self.offset = offset
        self.strike = strike

    @property
    def asian_definition(self):
        """
        Fixing details for asian options
        :return: object EtiOptionFixingInfo
        """
        return self._get_object_parameter(EtiFixingInfo, "asianDefinition")

    @asian_definition.setter
    def asian_definition(self, value):
        self._set_object_parameter(EtiFixingInfo, "asianDefinition", value)

    @property
    def barrier_definition(self):
        """
        Details for barrier option.
        :return: object EtiOptionBarrierDefinition
        """
        return self._get_object_parameter(EtiBarrierDefinition, "barrierDefinition")

    @barrier_definition.setter
    def barrier_definition(self, value):
        self._set_object_parameter(EtiBarrierDefinition, "barrierDefinition", value)

    @property
    def binary_definition(self):
        """
        Details for binary option.
        :return: object EtiOptionBinaryDefinition
        """
        return self._get_object_parameter(EtiBinaryDefinition, "binaryDefinition")

    @binary_definition.setter
    def binary_definition(self, value):
        self._set_object_parameter(EtiBinaryDefinition, "binaryDefinition", value)

    @property
    def buy_sell(self):
        """
        The side of the deal. Possible values:
        - Buy
        - Sell
        :return: enum BuySell
        """
        return self._get_enum_parameter(BuySell, "buySell")

    @buy_sell.setter
    def buy_sell(self, value):
        self._set_enum_parameter(BuySell, "buySell", value)

    @property
    def call_put(self):
        """
        Tells if the option is a call or a put. Possible values:
        - Call
        - Put
        :return: enum CallPut
        """
        return self._get_enum_parameter(CallPut, "callPut")

    @call_put.setter
    def call_put(self, value):
        self._set_enum_parameter(CallPut, "callPut", value)

    @property
    def cbbc_definition(self):
        """
        Details for CBBC (Call Bear/Bull Contract) option.
        :return: object EtiOptionCbbcDefinition
        """
        return self._get_object_parameter(EtiCbbcDefinition, "cbbcDefinition")

    @cbbc_definition.setter
    def cbbc_definition(self, value):
        self._set_object_parameter(EtiCbbcDefinition, "cbbcDefinition", value)

    @property
    def double_barriers_definition(self):
        """
        Details for double barriers option.
        :return: object EtiOptionDoubleBarriersDefinition
        """
        return self._get_object_parameter(
            EtiDoubleBarriersDefinition, "doubleBarriersDefinition"
        )

    @double_barriers_definition.setter
    def double_barriers_definition(self, value):
        self._set_object_parameter(
            EtiDoubleBarriersDefinition, "doubleBarriersDefinition", value
        )

    @property
    def exercise_style(self):
        """
        EURO or AMER
        :return: enum ExerciseStyle
        """
        return self._get_enum_parameter(ExerciseStyle, "exerciseStyle")

    @exercise_style.setter
    def exercise_style(self, value):
        self._set_enum_parameter(ExerciseStyle, "exerciseStyle", value)

    @property
    def underlying_definition(self):
        """
        Details of the underlying. Can be used to override some data of the underlying.
        :return: object EtiUnderlyingDefinition
        """
        return self._get_object_parameter(UnderlyingDefinition, "underlyingDefinition")

    @underlying_definition.setter
    def underlying_definition(self, value):
        self._set_object_parameter(UnderlyingDefinition, "underlyingDefinition", value)

    @property
    def underlying_type(self):
        """
        Underlying type of the option. Possible values:
        - Eti
        - Fx
        :return: enum UnderlyingType
        """
        return self._get_enum_parameter(UnderlyingType, "underlyingType")

    @underlying_type.setter
    def underlying_type(self, value):
        self._set_enum_parameter(UnderlyingType, "underlyingType", value)

    @property
    def deal_contract(self):
        """
        deal_contract. It is the number of contracts bought or sold in the deal.
        :return: int
        """
        return self._get_parameter("dealContract")

    @deal_contract.setter
    def deal_contract(self, value):
        self._set_parameter("dealContract", value)

    @property
    def end_date(self):
        """
        Expiry date of the option
        :return: str
        """
        return self._get_parameter("endDate")

    @end_date.setter
    def end_date(self, value):
        self._set_parameter("endDate", value)

    @property
    def end_date_time(self):
        """
        Expiry date time of the option
        :return: str
        """
        return self._get_parameter("endDateTime")

    @end_date_time.setter
    def end_date_time(self, value):
        self._set_parameter("endDateTime", value)

    @property
    def instrument_code(self):
        """
        An option RIC that is used to retrieve the description of the
        EtiOptionDefinition contract. Optional.If null, the instrument_code of
        underlying_definition must be provided.
        :return: str
        """
        return self._get_parameter("instrumentCode")

    @instrument_code.setter
    def instrument_code(self, value):
        self._set_parameter("instrumentCode", value)

    @property
    def lot_size(self):
        """
        The lot size. It is the number of options bought or sold in one transaction.
        :return: float
        """
        return self._get_parameter("lotSize")

    @lot_size.setter
    def lot_size(self, value):
        self._set_parameter("lotSize", value)

    @property
    def offset(self):
        """
        offset. The offset in minutes between the time UTC and the time of the exchange
        where the contract is traded.
        :return: int
        """
        return self._get_parameter("offset")

    @offset.setter
    def offset(self, value):
        self._set_parameter("offset", value)

    @property
    def strike(self):
        """
        strike of the option
        :return: float
        """
        return self._get_parameter("strike")

    @strike.setter
    def strike(self, value):
        self._set_parameter("strike", value)



import threading
from enum import Enum
from typing import Any, Callable, List, Optional, TYPE_CHECKING, Union

import numpy as np
import pandas as pd

if TYPE_CHECKING:
    from pandas import DataFrame


class RepeatedTimer(threading.Timer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.daemon = True

    def not_finished(self) -> bool:
        return not self.finished.is_set()

    def run(self):
        while self.not_finished():
            self.finished.wait(self.interval)
            if not self.finished.is_set():
                self.function(*self.args, **self.kwargs)


def ohlc(self, interval: str = "1min", *args, **kwargs) -> "DataFrame":
    call_from_recorder = kwargs.pop("call_from_recorder", False)
    df = self.fillna(method="ffill")
    df.fillna(method="bfill", inplace=True)
    df = df.astype("float")

    df = df.resample(interval, *args, **kwargs).ohlc()

    if call_from_recorder:
        df.fillna(pd.NA, inplace=True)
    return df


def get_from_path(obj, path, delim="."):
    if isinstance(path, list):
        splitted = path

    else:
        splitted = path.split(delim)

    for k in splitted:
        if hasattr(obj, "get"):
            obj = obj.get(k)

        elif iterable(obj) and is_int(k):
            try:
                obj = obj[int(k)]
            except (IndexError, KeyError):
                return None

    return obj


def is_int(obj):
    if isinstance(obj, str):
        try:
            int(obj)
        except Exception:
            return False
        else:
            return True
    return isinstance(obj, int)


def iterable(obj):
    if isinstance(obj, str):
        return False
    try:
        iter(obj)
    except TypeError:
        return False
    else:
        return True


def urljoin(*pieces):
    # first piece have a leading slash
    if pieces and len(pieces[0]) > 1 and pieces[0][0] == "/":
        pieces = ("/",) + pieces
    # last piece have a trailing slash
    if pieces and len(pieces[-1]) > 1 and pieces[-1][-1] == "/":
        pieces = pieces + ("/",)
    return "/".join(s.strip("/") for s in pieces)


def is_any_defined(*args):
    return any(args)


def is_all_defined(*args):
    return all(args)


def is_all_same_type(item_type, items):
    return all(isinstance(item, item_type) for item in items)


def make_counter():
    i = 0

    def counter():
        nonlocal i
        i += 1
        return i

    return counter


def get_response_reason(response):
    if hasattr(response, "reason_phrase"):
        return response.reason_phrase
    elif hasattr(response, "reason"):
        return response.reason
    return "unknown reason"


class cached_property(object):
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, cls=None):
        result = instance.__dict__[self.func.__name__] = self.func(instance)
        return result


def is_tuple(what):
    return type(what) == tuple


def is_all_str(input_list: list) -> bool:
    return all([isinstance(i, str) for i in input_list])


def validate_bool_value(value: bool) -> bool:
    if isinstance(value, bool):
        return value
    else:
        raise ValueError(
            f"Please provide boolean value ('True' or 'False'), "
            f"current value is '{value}'"
        )


def args_parser(param: Any, expect_none: bool = False) -> Optional[list]:
    if isinstance(param, str):
        return [param]

    if isinstance(param, list):
        if is_all_str(param):
            return param
        else:
            raise ValueError(f"Not all elements are strings in {param}")

    if param is None and expect_none:
        return None

    raise ValueError(f"Invalid type, expected str or list:\n {type(param)} is given")


class ArgsParser:
    def __init__(self, parse) -> None:
        self.parse = parse

    def get_str(self, *args, delim=None) -> str:
        if delim is not None:
            retval = delim.join(str(item) for item in self.get_list(*args))
        else:
            retval = self.parse(*args)
            if not isinstance(retval, str):
                retval = str(retval)
        return retval

    def get_list(self, *args) -> list:
        retval = self.parse(*args)
        if not isinstance(retval, list):
            retval = [retval]
        return retval

    def get_float(self, *args) -> float:
        retval = self.parse(*args)
        if isinstance(retval, np.datetime64):
            retval = retval.astype(float)
        else:
            retval = float(retval)
        return retval

    def get_bool(self, *args) -> bool:
        retval = self.parse(*args)
        if not isinstance(retval, bool):
            retval = bool(retval)
        return retval


def parse_list_of_str(param: Union[str, list]) -> list:
    if isinstance(param, str):
        return [param]

    if isinstance(param, list):
        if is_all_same_type(str, param):
            return param
        else:
            raise ValueError(f"Not all elements are strings in {param}")

    raise TypeError(
        f"Invalid type, expected str or list, {type(param).__name__} is given"
    )


def make_parse_enum(enum_class: iterable, can_be_lower: bool = False) -> callable:
    enum_values = [k.value for k in enum_class]

    def parse_enum(param: Union[str, list, Enum]) -> Union[str, list]:
        if isinstance(param, list):
            return [parse_enum(p) for p in param]

        if isinstance(param, enum_class):
            return param.value

        if param in enum_values:
            return param

        if can_be_lower and hasattr(param, "upper"):
            param_upper = param.upper()
            if param_upper in enum_values:
                return param_upper

        raise AttributeError(f"Value '{param}' must be in {enum_values}")

    return parse_enum


def make_convert_to_enum(enum_class: iterable) -> callable:
    def convert_to_enum(
        param: Union[str, List[str], Enum, List[Enum]]
    ) -> Union[Enum, List[Enum]]:
        if isinstance(param, list):
            return [convert_to_enum(p) for p in param]

        if isinstance(param, enum_class):
            return param

        if isinstance(param, str) and param.upper() in enum_class.__members__.keys():
            param = enum_class.__members__[param.upper()]
            return param

        raise ValueError(f"Cannot convert param '{param}'")

    return convert_to_enum


def make_callback(func: Callable) -> Callable:
    """Return a callback function with correct arguments order.

    For user defined callback functions in OMMStream, RDPStream and trade data streams,
    correct arguments order is:
        Callable[[dict, "Stream"], Any]

    For user defined callback functions in pricing streams correct arguments order is:
        Callable[[Any, str, "Stream"], Any]

    For user defined callback in financial contracts correct arguments order is:
        Callable[[list, "Stream"], Any]

    Uses in OMMStream, RDPStream, financial contracts,  pricing and trade data streams.
    """

    def callback(*args):
        args = reversed(args)
        func(*args)

    return callback


def parse_hp_universe(universe: Union[str, list]) -> list:
    universe = parse_list_of_str(universe)
    universe = [i for i in list(dict.fromkeys(universe)) if i not in {"", " "}]
    if universe:
        return universe
    raise ValueError("List of universes is empty, nothing to process")


def validate_types(val: Any, types: Union[list, tuple], val_name: str = "") -> None:
    t_names = [tp.__name__ for tp in types if tp if tp.__name__ != "NoneType"]

    if None in types:
        raise TypeError("Use 'type(None)' instead 'None', in 'types'")

    if type(val) not in types:
        raise TypeError(
            f"Parameter '{val_name}' of invalid type provided: '{type(val).__name__}', "
            f"expected types: {t_names}"
        )


fields_arg_parser = ArgsParser(parse_list_of_str)
universe_arg_parser = fields_arg_parser
hp_universe_parser = ArgsParser(parse_hp_universe)
custom_insts_historical_universe_parser = ArgsParser(parse_hp_universe)


def make_enum_arg_parser(*args, **kwargs):
    return ArgsParser(make_parse_enum(*args, **kwargs))


def make_convert_to_enum_arg_parser(*args, **kwargs):
    return ArgsParser(make_convert_to_enum(*args, **kwargs))



def do_all_elements_have_error(data: dict, elements: list) -> bool:
    counter = len(elements) or 1
    for element in elements:

        if not hasattr(element, "get"):
            counter -= 1
            data["error_message"] = f"Invalid data type={type(element)}, data={element}"
            continue

        error = element.get("error")

        if error:
            counter -= 1
            error_codes = data.setdefault("error_code", [])
            error_code = error.get("code")
            error_codes.append(error_code)
            error_messages = data.setdefault("error_message", [])
            error_message = error.get("message")
            error_messages.append(error_message)

    if counter == 0:
        return True

    return False



# coding: utf8

from typing import Optional

from ._repo_underlying_pricing_parameters import UnderlyingPricingParameters
from .._instrument_definition import InstrumentDefinition
from ..._object_definition import ObjectDefinition


class UnderlyingContract(ObjectDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    instrument_type : str, optional
        The type of instrument being defined.
    instrument_definition : object, optional
        Definition of the input contract
    pricing_parameters : UnderlyingPricingParameters, optional
        The pricing parameters to apply to this instrument. If pricing parameters are
        not provided at this level parameters defined globally at the request level are
        used. If no pricing parameters are provided globally default values apply.
    csa_tag : str, optional

    Examples
    --------
     >>> import refinitiv.data.content.ipa.financial_contracts as rdf
     >>> rdf.repo.UnderlyingContract(
     ...    instrument_type="Bond",
     ...    instrument_definition=rdf.bond.Definition(instrument_code="US191450264="),
     ...)
    """

    def __init__(
        self,
        instrument_type: Optional[str] = None,
        instrument_definition: Optional[object] = None,
        pricing_parameters: Optional[UnderlyingPricingParameters] = None,
        csa_tag: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.instrument_type = instrument_type
        self.instrument_definition = instrument_definition
        self.pricing_parameters = pricing_parameters
        self.csa_tag = csa_tag

    @property
    def instrument_definition(self):
        """
        Definition of the input contract
        :return: object InstrumentDefinition
        """
        return self._get_object_parameter(InstrumentDefinition, "instrumentDefinition")

    @instrument_definition.setter
    def instrument_definition(self, value):
        self._set_object_parameter(InstrumentDefinition, "instrumentDefinition", value)

    @property
    def pricing_parameters(self):
        """
        The pricing parameters to apply to this instrument. Optional.
        If pricing parameters are not provided at this level parameters defined globally at the request level are used. If no
        pricing parameters are provided globally default values apply.
        :return: object RepoUnderlyingPricingParameters
        """
        return self._get_object_parameter(
            UnderlyingPricingParameters, "pricingParameters"
        )

    @pricing_parameters.setter
    def pricing_parameters(self, value):
        self._set_object_parameter(
            UnderlyingPricingParameters, "pricingParameters", value
        )

    @property
    def csa_tag(self):
        """
        :return: str
        """
        return self._get_parameter("csaTag")

    @csa_tag.setter
    def csa_tag(self, value):
        self._set_parameter("csaTag", value)

    @property
    def instrument_type(self):
        """
        The type of instrument being defined.
        :return: str
        """
        return self._get_parameter("instrumentType")

    @instrument_type.setter
    def instrument_type(self, value):
        self._set_parameter("instrumentType", value)



# coding: utf8

__all__ = ["EquityDividendType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class EquityDividendType(Enum):
    DEFAULT = "Default"
    DISCRETE = "Discrete"
    YIELD = "Yield"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(EquityDividendType, _EQUITY_DIVIDEND_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(
            _EQUITY_DIVIDEND_TYPE_VALUES_IN_LOWER_BY_EQUITY_DIVIDEND_TYPE, some
        )


_EQUITY_DIVIDEND_TYPE_VALUES = tuple(t.value for t in EquityDividendType)
_EQUITY_DIVIDEND_TYPE_VALUES_IN_LOWER_BY_EQUITY_DIVIDEND_TYPE = {
    name.lower(): item for name, item in EquityDividendType.__members__.items()
}



# coding: utf8
# contract_gen 2020-09-02 06:21:29.309974

__all__ = ["IndexConvexityAdjustmentMethod"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class IndexConvexityAdjustmentMethod(Enum):
    NONE = "None"
    BLACK_SCHOLES = "BlackScholes"
    REPLICATION = "Replication"
    LINEAR_SWAP_MODEL = "LinearSwapModel"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            IndexConvexityAdjustmentMethod,
            _INDEX_CONVEXITY_ADJUSTMENT_METHOD_VALUES,
            some,
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _INDEX_CONVEXITY_ADJUSTMENT_METHOD_VALUES_IN_LOWER_BY_INDEX_CONVEXITY_ADJUSTMENT_METHOD,
            some,
        )


_INDEX_CONVEXITY_ADJUSTMENT_METHOD_VALUES = tuple(
    t.value for t in IndexConvexityAdjustmentMethod
)
_INDEX_CONVEXITY_ADJUSTMENT_METHOD_VALUES_IN_LOWER_BY_INDEX_CONVEXITY_ADJUSTMENT_METHOD = {
    name.lower(): item
    for name, item in IndexConvexityAdjustmentMethod.__members__.items()
}



from typing import Optional

from .._content_type import ContentType
from ._data_provider_layer import (
    CustomInstrumentDataProviderLayer,
)


class Definition(CustomInstrumentDataProviderLayer):
    """
    Allows to create custom instrument, symbol and formula fields are mandatory, others are optional

    Parameters
    -----------
    universe : str
        The Id or Symbol of custom instrument to operate on. Use only for get_instrument().

    Examples
    --------
    >>> from refinitiv.data.content.custom_instruments import Definition
    >>> definition = Definition(universe="MyNewInstrument")
    >>> stream = definition.get_stream()
    """

    def __init__(
        self,
        universe: Optional[str],
    ):
        super().__init__(
            data_type=ContentType.CUSTOM_INSTRUMENTS_INSTRUMENTS,
            universe=universe,
        )



# coding: utf-8


class Grant(object):
    def __init__(self, *args, **kwargs):
        if len(args):
            self._username = args[0]
        elif kwargs.get("username"):
            self._username = kwargs.get("username")
        else:
            self._username = None

    def get_username(self):
        return self._username

    def username(self, user_name):
        self._username = user_name
        return self



from enum import Enum, unique


@unique
class AssetState(Enum):
    """
    Asset state values for 'filter' parameter in request for SymbolConversion content object.
    """

    ACTIVE = "Active"
    INACTIVE = "Inactive"



# coding: utf8

from enum import Enum, unique

from ..._enums._common_tools import _convert_to_str, _normalize


@unique
class SwapPriceSide(Enum):
    ASK = "Ask"
    BID = "Bid"
    MID = "Mid"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(SwapPriceSide, _SWAP_PRICE_SIDE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_SWAP_PRICE_SIDE_VALUES_IN_LOWER_BY_SWAP_PRICE_SIDE, some)


_SWAP_PRICE_SIDE_VALUES = tuple(t.value for t in SwapPriceSide)
_SWAP_PRICE_SIDE_VALUES_IN_LOWER_BY_SWAP_PRICE_SIDE = {
    name.lower(): item for name, item in SwapPriceSide.__members__.items()
}



# coding: utf8

from .._object_definition import ObjectDefinition


class EtiSurfaceDefinition(ObjectDefinition):
    def __init__(
        self,
        instrument_code=None,
        clean_instrument_code=None,
        exchange=None,
        is_future_underlying=None,
    ):
        super().__init__()
        self.instrument_code = instrument_code
        self.clean_instrument_code = clean_instrument_code
        self.exchange = exchange
        self.is_future_underlying = is_future_underlying

    @property
    def clean_instrument_code(self):
        """
        :return: str
        """
        return self._get_parameter("cleanInstrumentCode")

    @clean_instrument_code.setter
    def clean_instrument_code(self, value):
        self._set_parameter("cleanInstrumentCode", value)

    @property
    def exchange(self):
        """
        Specifies the exchange to be used to retrieve the underlying data.
        :return: str
        """
        return self._get_parameter("exchange")

    @exchange.setter
    def exchange(self, value):
        self._set_parameter("exchange", value)

    @property
    def instrument_code(self):
        """
        The code (RIC for equities and indices and RICROOT for Futures.) that represents the instrument.
        The format for equities and indices is xxx@RIC (Example: VOD.L@RIC)
        The format for Futures is xx@RICROOT (Example: CL@RICROOT)
        :return: str
        """
        return self._get_parameter("instrumentCode")

    @instrument_code.setter
    def instrument_code(self, value):
        self._set_parameter("instrumentCode", value)

    @property
    def is_future_underlying(self):
        """
        :return: bool
        """
        return self._get_parameter("isFutureUnderlying")

    @is_future_underlying.setter
    def is_future_underlying(self, value):
        self._set_parameter("isFutureUnderlying", value)



# coding: utf8
# contract_gen 2020-09-02 06:21:29.308976

__all__ = ["IndexConvexityAdjustmentIntegrationMethod"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class IndexConvexityAdjustmentIntegrationMethod(Enum):
    RIEMANN_SUM = "RiemannSum"
    RUNGE_KUTTA = "RungeKutta"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            IndexConvexityAdjustmentIntegrationMethod,
            _INDEX_CONVEXITY_ADJUSTMENT_INTEGRATION_METHOD_VALUES,
            some,
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _INDEX_CONVEXITY_ADJUSTMENT_INTEGRATION_METHOD_VALUES_IN_LOWER_BY_INDEX_CONVEXITY_ADJUSTMENT_INTEGRATION_METHOD,
            some,
        )


_INDEX_CONVEXITY_ADJUSTMENT_INTEGRATION_METHOD_VALUES = tuple(
    t.value for t in IndexConvexityAdjustmentIntegrationMethod
)
_INDEX_CONVEXITY_ADJUSTMENT_INTEGRATION_METHOD_VALUES_IN_LOWER_BY_INDEX_CONVEXITY_ADJUSTMENT_INTEGRATION_METHOD = {
    name.lower(): item
    for name, item in IndexConvexityAdjustmentIntegrationMethod.__members__.items()
}



from .._data._data_type import DataType
from ._data_class import AttributesCFS
from ._data_types import BucketData, FileData, FileSetData, PackageData
from ._base_definition import BaseDefinition
from ..._tools._utils import camel_to_snake


class BaseCFSObject(AttributesCFS):
    def __init__(self, data, provider, session=None):
        self._data = {}
        self._session = session
        super().__init__(provider, self._data)
        self._data.update({camel_to_snake(key): value for key, value in data.items()})
        self._child_objects = None
        self._provider = provider
        for name, value in self._data.items():
            setattr(self, name, value)

    def __getitem__(self, item):
        return self._data[item]

    def __iter__(self):
        if self._child_type is None:
            raise TypeError(f"object {self._provider.name, type(self)} is not iterable")

        args = self._params
        kwargs = {args[0]: self._data[args[1]]}
        bd = BaseDefinition(data_type=self._child_type, **kwargs)

        self._child_objects = bd.get_data(session=self._session)
        self._n = 0
        return self

    def __next__(self):
        _iter_obj = self._child_objects.data._iter_object
        if not _iter_obj:
            raise StopIteration
        if self._n < len(_iter_obj):
            result = self._child_objects.data._iter_object[self._n]
            self._n += 1
            return result
        raise StopIteration

    def __repr__(self):
        return f"{self.__class__.__name__}({self._data})"

    @property
    def _params(self):
        raise NotImplementedError

    @property
    def _child_type(self):
        raise NotImplementedError


class CFSFile(BaseCFSObject):
    _child_type = None
    _params = None


class CFSBucket(BaseCFSObject):
    _child_type = DataType.CFS_FILE_SETS
    _params = ("bucket", "name")


class CFSFileSet(BaseCFSObject):
    _child_type = DataType.CFS_FILES
    _params = ("fileset_id", "id")


class CFSPackage(BaseCFSObject):
    _child_type = DataType.CFS_FILE_SETS
    _params = ("package_id", "package_id")

    def __init__(self, data, provider, session=None):
        super().__init__(data, provider, session=session)
        self._bucket_names = self._data["bucket_names"]

    def __iter__(self):
        args = self._params
        kwargs = {args[0]: self._data[args[1]]}
        bucket_name = self._bucket_names[0] if self._bucket_names else []
        bd = BaseDefinition(
            data_type=self._child_type,
            bucket=bucket_name,
            **kwargs,
        )
        self._child_objects = bd.get_data(session=self._session)
        self._n = 0
        return self

    def __next__(self):
        _iter_obj = self._child_objects.data._iter_object
        if not _iter_obj:
            raise StopIteration
        if self._n < len(_iter_obj):
            result = self._child_objects.data._iter_object[self._n]
            self._n += 1
            return result
        while self._bucket_names:
            bucket = self._bucket_names.pop(0)
            args = self._params
            kwargs = {args[0]: self._data[args[1]]}
            bd = BaseDefinition(
                data_type=self._child_type,
                bucket=bucket,
                **kwargs,
            )
            self._child_objects = bd.get_data()
            if not self._child_objects.errors and self._child_objects.data.raw["value"]:
                self._n = 0
                result = self._child_objects.data._iter_object[self._n]
                self._n += 1
                return result
        raise StopIteration


class_by_type = {
    BucketData: CFSBucket,
    PackageData: CFSPackage,
    FileSetData: CFSFileSet,
    FileData: CFSFile,
}


class IterObj:
    def __init__(self, value, session=None, provider=None):
        _class = class_by_type.get(provider)
        self._values = [_class(i, provider=provider, session=session) for i in value]

    def __getitem__(self, item):
        return self._values[item]

    def __iter__(self):
        self._n = 0
        return self

    def __next__(self):
        if self._n < len(self._values):
            result = self._values[self._n]
            self._n += 1
            return result
        raise StopIteration

    def __repr__(self):
        return "\n".join([repr(i) for i in self._values])

    def __len__(self):
        return len(self._values)



import json
import time
import logging
from typing import Callable, Optional

try:
    from ema import (
        OmmConsumer,
        OmmConsumerConfig,
        AppClient,
        CustomLoggerClient,
        LoggerSeverity,
        MapEntry,
    )  # noqa

    EMA_INSTALLED = True
except ImportError:
    EMA_INSTALLED = False
else:
    from .conversion import json_marketprice_msg_to_ema
    from .ema import (
        ema_login_message,
        create_programmatic_cfg,
        name_type_map,
        generate_login_msg,
    )


if EMA_INSTALLED:

    class EmaPythonLogger(CustomLoggerClient):
        def __init__(self):
            super().__init__(EmaPythonLogger.log)

        severity_logging_map: dict = {
            LoggerSeverity.NoLogMsg: 1000,  # higher than critical to avoid logging
            LoggerSeverity.Error: logging.ERROR,
            LoggerSeverity.Warning: logging.WARNING,
            LoggerSeverity.Success: logging.INFO,
            LoggerSeverity.Verbose: logging.DEBUG,
        }

        @staticmethod
        def log(callback_client_name: str, severity: LoggerSeverity, message: str):
            logging.log(
                EmaPythonLogger.severity_logging_map[severity],
                f"PythonLogger - {callback_client_name}: {message}",
            )


class RwfSocketApp:
    """Emulating WebsocketApp + WSAPI"""

    def __init__(
        self,
        host: str,
        port: int,
        on_open: Callable,
        on_msg: Callable,
        on_close: Callable,
        field_dict_path: Optional[str] = None,
        enumtype_path: Optional[str] = None,
        python_logger: bool = True,
    ):
        if not EMA_INSTALLED:
            raise ImportError("You need to install refinitiv-ema to use RwfSocketApp")

        self.host = host
        self.port = port

        self.field_dict_path = field_dict_path
        self.enumtype_path = enumtype_path
        self.on_open = on_open
        self.on_msg = on_msg
        self.on_close = on_close
        self.id = None
        self.handles = {}  # stream_id: handle
        self.consumer = None
        self.keep_running = False
        self._skip_login = True
        self._reissue_handle = None

        def refresh_msg_cb(msg, event):
            # TODO: Ema example has LoginRefresh and reissue timestamp inside,
            #  possibly useful?
            self._reissue_handle = event.get_handle()

        def msg_cb(msg, event):
            on_msg(self, f"[{msg.json()}]")

        def nothing_callback(msg, event):
            pass
            # logger.debug(f"login client message: {msg.json()}")

        self.client = AppClient(
            on_refresh_msg=msg_cb,
            on_update_msg=msg_cb,
            on_status_msg=msg_cb,
        )

        self.login_client = AppClient(
            on_refresh_msg=refresh_msg_cb,
            on_update_msg=nothing_callback,
            on_status_msg=nothing_callback,
        )
        if python_logger:
            self.ema_logger = EmaPythonLogger()
        else:
            self.ema_logger = None  # Will use internal OmmLoggerClient

    def close(self, **_):
        self.keep_running = False
        self.on_close(self, -100, "Just Closed")

    def run_forever(self, login_msg: dict):
        admin_msg = ema_login_message(**generate_login_msg(login_msg))
        pgcfg = create_programmatic_cfg(
            field_dict_path=self.field_dict_path,
            enumtype_path=self.enumtype_path,
            host=self.host,
            port=self.port,
        )
        config = OmmConsumerConfig().config(pgcfg).add_admin_msg(admin_msg)

        self.consumer: OmmConsumer = OmmConsumer(
            config, self.login_client, self.ema_logger
        )
        self.on_open(self)
        self.keep_running = True

        try:
            while self.keep_running:
                time.sleep(0.2)

        except (Exception, KeyboardInterrupt, SystemExit) as e:
            # self._callback(self.on_error, e)
            if isinstance(e, SystemExit):
                # propagate SystemExit further
                raise
            # teardown()
            return not isinstance(e, KeyboardInterrupt)

    def send(self, msg: str):
        msg = json.loads(msg)

        msg_type = msg.get("Type", "Request")
        msg_domain = msg.get("Domain", "MarketPrice")

        if msg_type == "Pong":
            # do nothing
            pass

        elif msg_type == "Request":  # open
            if msg_domain == "MarketPrice":
                ema_msg = json_marketprice_msg_to_ema(msg, self.consumer.field_id_map)
                handle = self.consumer.register_client(ema_msg, self.client)
                # TODO What if we have wrong message id?
                self.handles[msg["ID"]] = handle
                # logger.debug(f"new client registered for {msg['Key']['Name']}")
            elif msg_domain == "Login":
                if self._skip_login:
                    self._skip_login = False
                    # logger.debug("First login message sent during init, skipping...")
                elif self._reissue_handle is not None:
                    admin_msg = ema_login_message(**generate_login_msg(msg))
                    self.consumer.reissue(admin_msg, self._reissue_handle)
                else:
                    # do nothing
                    pass
                    # logger.error(
                    #     "Tried re-login, but has not received valid login "
                    #     "response beforehand."
                    # )

            else:
                raise ValueError(f"Unknown domain of Request message: {msg_domain}")

        elif msg_type == "Close":
            if msg_domain == "Login":
                self.close()  # TODO: is this right?
            else:
                # logger.debug("RwfSocket received Close message")
                self.consumer.unregister(self.handles[msg["ID"]])

        else:
            raise ValueError(f"Unknown message type: {msg_type}")



# coding: utf8

from typing import Optional

from .._base import UnderlyingDefinition
from ..._instrument_definition import InstrumentDefinition
from .._enums import (
    BuySell,
    CallPut,
    ExerciseStyle,
    UnderlyingType,
)
from . import (
    FxDualCurrencyDefinition,
    FxAverageInfo,
    FxBarrierDefinition,
    FxBinaryDefinition,
    FxDoubleBarrierDefinition,
    FxDoubleBinaryDefinition,
    FxForwardStart,
    FxUnderlyingDefinition,
)


class FxDefinition(InstrumentDefinition):
    """
    Parameters
    ----------
    instrument_tag : str, optional
        User defined string to identify the instrument.It can be used to link output
        results to the instrument definition. Only alphabetic, numeric and '- _.#=@'
        characters are supported. Optional.
    end_date : str, optional
        Expiry date of the option
    tenor : str, optional
        tenor of the option
    notional_ccy : str, optional
        Currency of the notional amount If the option is a EURGBP Call option,
        notional_ccy can be expressed in EUR OR GBP
    notional_amount : float, optional
        The notional amount of currency If the option is a EURGBP Call option, amount of
        EUR or GBP of the contract
    asian_definition : FxOptionAverageInfo, optional
        Fixing details for asian options
    barrier_definition : FxOptionBarrierDefinition, optional
        Details for barrier option.
    binary_definition : FxOptionBinaryDefinition, optional
        Details for binary option.
    buy_sell : BuySell, optional
        The side of the deal. Possible values:
        - Buy
        - Sell
    call_put : CallPut, optional
        Tells if the option is a call or a put. Possible values:
        - Call
        - Put
    double_barrier_definition : FxOptionDoubleBarrierDefinition, optional
        Details for double barriers option.
    double_binary_definition : FxOptionDoubleBinaryDefinition, optional
        Details for double binary option.
    dual_currency_definition : FxDualCurrencyDefinition, optional
        Details for dual currency option.
    exercise_style : ExerciseStyle, optional
        EURO or AMER
    forward_start_definition : FxOptionForwardStart, optional
        Details for Forward Start option.
    underlying_definition : FxUnderlyingDefinition, optional
        The details of the underlying
    underlying_type : UnderlyingType, optional
        Underlying type of the option. Possible values:
        - Eti
        - Fx
    delivery_date : str, optional
        Expiry date of the option
    strike : float, optional
        strike of the option
    """

    def __init__(
        self,
        instrument_tag: Optional[str] = None,
        end_date: Optional[str] = None,
        tenor: Optional[str] = None,
        notional_ccy: Optional[str] = None,
        notional_amount: Optional[float] = None,
        asian_definition: Optional[FxAverageInfo] = None,
        barrier_definition: Optional[FxBarrierDefinition] = None,
        binary_definition: Optional[FxBinaryDefinition] = None,
        buy_sell: Optional[BuySell] = None,
        call_put: Optional[CallPut] = None,
        double_barrier_definition: Optional[FxDoubleBarrierDefinition] = None,
        double_binary_definition: Optional[FxDoubleBinaryDefinition] = None,
        dual_currency_definition: Optional[FxDualCurrencyDefinition] = None,
        exercise_style: Optional[ExerciseStyle] = None,
        forward_start_definition: Optional[FxForwardStart] = None,
        underlying_definition: Optional[FxUnderlyingDefinition] = None,
        underlying_type: Optional[UnderlyingType] = None,
        delivery_date: Optional[str] = None,
        strike: Optional[float] = None,
        **kwargs,
    ) -> None:
        super().__init__(instrument_tag, **kwargs)
        self.end_date = end_date
        self.tenor = tenor
        self.notional_ccy = notional_ccy
        self.notional_amount = notional_amount
        self.asian_definition = asian_definition
        self.barrier_definition = barrier_definition
        self.binary_definition = binary_definition
        self.buy_sell = buy_sell
        self.call_put = call_put
        self.double_barrier_definition = double_barrier_definition
        self.double_binary_definition = double_binary_definition
        self.dual_currency_definition = dual_currency_definition
        self.exercise_style = exercise_style
        self.forward_start_definition = forward_start_definition
        self.underlying_definition = underlying_definition
        self.underlying_type = underlying_type
        self.delivery_date = delivery_date
        self.strike = strike

    @property
    def asian_definition(self):
        """
        Fixing details for asian options
        :return: object FxOptionAverageInfo
        """
        return self._get_object_parameter(FxAverageInfo, "asianDefinition")

    @asian_definition.setter
    def asian_definition(self, value):
        self._set_object_parameter(FxAverageInfo, "asianDefinition", value)

    @property
    def barrier_definition(self):
        """
        Details for barrier option.
        :return: object FxOptionBarrierDefinition
        """
        return self._get_object_parameter(FxBarrierDefinition, "barrierDefinition")

    @barrier_definition.setter
    def barrier_definition(self, value):
        self._set_object_parameter(FxBarrierDefinition, "barrierDefinition", value)

    @property
    def binary_definition(self):
        """
        Details for binary option.
        :return: object FxOptionBinaryDefinition
        """
        return self._get_object_parameter(FxBinaryDefinition, "binaryDefinition")

    @binary_definition.setter
    def binary_definition(self, value):
        self._set_object_parameter(FxBinaryDefinition, "binaryDefinition", value)

    @property
    def buy_sell(self):
        """
        The side of the deal. Possible values:
        - Buy
        - Sell
        :return: enum BuySell
        """
        return self._get_enum_parameter(BuySell, "buySell")

    @buy_sell.setter
    def buy_sell(self, value):
        self._set_enum_parameter(BuySell, "buySell", value)

    @property
    def call_put(self):
        """
        Tells if the option is a call or a put. Possible values:
        - Call
        - Put
        :return: enum CallPut
        """
        return self._get_enum_parameter(CallPut, "callPut")

    @call_put.setter
    def call_put(self, value):
        self._set_enum_parameter(CallPut, "callPut", value)

    @property
    def double_barrier_definition(self):
        """
        Details for double barriers option.
        :return: object FxOptionDoubleBarrierDefinition
        """
        return self._get_object_parameter(
            FxDoubleBarrierDefinition, "doubleBarrierDefinition"
        )

    @double_barrier_definition.setter
    def double_barrier_definition(self, value):
        self._set_object_parameter(
            FxDoubleBarrierDefinition, "doubleBarrierDefinition", value
        )

    @property
    def double_binary_definition(self):
        """
        Details for double binary option.
        :return: object FxOptionDoubleBinaryDefinition
        """
        return self._get_object_parameter(
            FxDoubleBinaryDefinition, "doubleBinaryDefinition"
        )

    @double_binary_definition.setter
    def double_binary_definition(self, value):
        self._set_object_parameter(
            FxDoubleBinaryDefinition, "doubleBinaryDefinition", value
        )

    @property
    def dual_currency_definition(self):
        """
        Details for dual currency option.
        :return: object FxDualCurrencyDefinition
        """
        return self._get_object_parameter(
            FxDualCurrencyDefinition, "dualCurrencyDefinition"
        )

    @dual_currency_definition.setter
    def dual_currency_definition(self, value):
        self._set_object_parameter(
            FxDualCurrencyDefinition, "dualCurrencyDefinition", value
        )

    @property
    def exercise_style(self):
        """
        EURO or AMER
        :return: enum ExerciseStyle
        """
        return self._get_enum_parameter(ExerciseStyle, "exerciseStyle")

    @exercise_style.setter
    def exercise_style(self, value):
        self._set_enum_parameter(ExerciseStyle, "exerciseStyle", value)

    @property
    def forward_start_definition(self):
        """
        Details for Forward Start option.
        :return: object FxOptionForwardStart
        """
        return self._get_object_parameter(FxForwardStart, "forwardStartDefinition")

    @forward_start_definition.setter
    def forward_start_definition(self, value):
        self._set_object_parameter(FxForwardStart, "forwardStartDefinition", value)

    @property
    def underlying_definition(self):
        """
        The details of the underlying
        :return: object FxUnderlyingDefinition
        """
        return self._get_object_parameter(UnderlyingDefinition, "underlyingDefinition")

    @underlying_definition.setter
    def underlying_definition(self, value):
        self._set_object_parameter(UnderlyingDefinition, "underlyingDefinition", value)

    @property
    def underlying_type(self):
        """
        Underlying type of the option. Possible values:
        - Eti
        - Fx
        :return: enum UnderlyingType
        """
        return self._get_enum_parameter(UnderlyingType, "underlyingType")

    @underlying_type.setter
    def underlying_type(self, value):
        self._set_enum_parameter(UnderlyingType, "underlyingType", value)

    @property
    def delivery_date(self):
        """
        Expiry date of the option
        :return: str
        """
        return self._get_parameter("deliveryDate")

    @delivery_date.setter
    def delivery_date(self, value):
        self._set_parameter("deliveryDate", value)

    @property
    def end_date(self):
        """
        Expiry date of the option
        :return: str
        """
        return self._get_parameter("endDate")

    @end_date.setter
    def end_date(self, value):
        self._set_parameter("endDate", value)

    @property
    def notional_amount(self):
        """
        The notional amount of currency If the option is a EURGBP Call option, amount of
        EUR or GBP of the contract
        :return: float
        """
        return self._get_parameter("notionalAmount")

    @notional_amount.setter
    def notional_amount(self, value):
        self._set_parameter("notionalAmount", value)

    @property
    def notional_ccy(self):
        """
        Cuurency of the notional amount If the option is a EURGBP Call option,
        notional_ccy can be expressed in EUR OR GBP
        :return: str
        """
        return self._get_parameter("notionalCcy")

    @notional_ccy.setter
    def notional_ccy(self, value):
        self._set_parameter("notionalCcy", value)

    @property
    def strike(self):
        """
        strike of the option
        :return: float
        """
        return self._get_parameter("strike")

    @strike.setter
    def strike(self, value):
        self._set_parameter("strike", value)

    @property
    def tenor(self):
        """
        tenor of the option
        :return: str
        """
        return self._get_parameter("tenor")

    @tenor.setter
    def tenor(self, value):
        self._set_parameter("tenor", value)



def _normalize(some: str, symbols_type: dict) -> str:
    some_lower = some.lower()
    symbol_type = symbols_type.get(some_lower)
    result = ""
    if symbol_type:
        result = symbol_type.value
    return result



from datetime import date, datetime, timedelta
from typing import Optional, Union

from ._data_provider import get_content_type_by_interval
from .._intervals import DayIntervalType, get_day_interval_type, Intervals
from ..._tools import validate_types, custom_insts_historical_universe_parser
from ...delivery._data._data_provider import DataProviderLayer


class Definition(DataProviderLayer):
    """
    Summary line of this class that defines parameters for requesting summaries from custom instruments

    Parameters
    ----------
    universe : str or list
        The Id or Symbol of custom instrument to operate on
    interval : str or Intervals, optional
        The consolidation interval in ISO8601
    start : str or date or datetime or timedelta, optional
        The start date and timestamp of the query in ISO8601 with UTC only
    end : str or date or datetime or timedelta, optional
        The end date and timestamp of the query in ISO8601 with UTC only
    count : int, optional
        The maximum number of data returned. Values range: 1 - 10000

    Examples
    --------
    >>> from refinitiv.data.content.custom_instruments import summaries
    >>> definition_summaries = summaries.Definition("VOD.L")
    >>> response = definition_summaries.get_data()
    """

    def __init__(
        self,
        universe: Union[str, list],
        interval: Union[str, Intervals] = None,
        start: Optional[Union[str, date, datetime, timedelta]] = None,
        end: Optional[Union[str, date, datetime, timedelta]] = None,
        count: Optional[int] = None,
    ) -> None:
        day_interval_type = get_day_interval_type(interval or DayIntervalType.INTER)
        content_type = get_content_type_by_interval(day_interval_type)
        validate_types(count, [int, type(None)], "count")

        universe = custom_insts_historical_universe_parser.get_list(universe)

        super().__init__(
            data_type=content_type,
            day_interval_type=day_interval_type,
            universe=universe,
            interval=interval,
            start=start,
            end=end,
            count=count,
        )



# coding: utf8
# contract_gen 2020-05-18 08:30:59.242815

__all__ = ["FundingSpreadMethod"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class FundingSpreadMethod(Enum):
    CDS_SPREAD = "CDSSpread"
    LIQUIDITY_SPREAD = "LiquiditySpread"
    PPZ_SPREAD = "PPZSpread"
    PARALLEL_CURVE_SHIFT = "ParallelCurveShift"
    RISKY_D_FS = "RiskyDFs"
    USE_CDS_SPREADS = "UseCDSSpreads"
    USE_Z_SPREAD = "UseZSpread"
    Z_SPREAD = "ZSpread"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(FundingSpreadMethod, _FUNDING_SPREAD_METHOD_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(
            _FUNDING_SPREAD_METHOD_VALUES_IN_LOWER_BY_FUNDING_SPREAD_METHOD, some
        )


_FUNDING_SPREAD_METHOD_VALUES = tuple(t.value for t in FundingSpreadMethod)
_FUNDING_SPREAD_METHOD_VALUES_IN_LOWER_BY_FUNDING_SPREAD_METHOD = {
    name.lower(): item for name, item in FundingSpreadMethod.__members__.items()
}



# coding: utf8
# contract_gen 2020-05-18 08:30:59.210816

__all__ = ["RedemptionDateType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class RedemptionDateType(Enum):
    """
    Redemption type of the bond.

    It is used to compute the default redemption date:
        - RedemptionAtMaturityDate : yield and price are computed at maturity date.
        - RedemptionAtCallDate : yield and price are computed at call date (next call
          date by default).
        - RedemptionAtPutDate : yield and price are computed at put date (next put date
          by default)..
        - RedemptionAtWorstDate : yield and price are computed at the lowest yield date.
        - RedemptionAtSinkDate : yield and price are computed at sinking fund date.
        - RedemptionAtParDate : yield and price are computed at next par.
        - RedemptionAtPremiumDate : yield and price are computed at next premium.
        - RedemptionAtMakeWholeCallDate : yield and price are computed at Make Whole
          Call date.
        - RedemptionAtAverageLife : yield and price are computed at average life (case
          of sinkable bonds)
        - RedemptionAtNextDate : yield and price are computed at next redemption date
          available.
    """

    REDEMPTION_AT_AVERAGE_LIFE = "RedemptionAtAverageLife"
    REDEMPTION_AT_BEST_DATE = "RedemptionAtBestDate"
    REDEMPTION_AT_CALL_DATE = "RedemptionAtCallDate"
    REDEMPTION_AT_CUSTOM_DATE = "RedemptionAtCustomDate"
    REDEMPTION_AT_MAKE_WHOLE_CALL_DATE = "RedemptionAtMakeWholeCallDate"
    REDEMPTION_AT_MATURITY_DATE = "RedemptionAtMaturityDate"
    REDEMPTION_AT_NEXT_DATE = "RedemptionAtNextDate"
    REDEMPTION_AT_PAR_DATE = "RedemptionAtParDate"
    REDEMPTION_AT_PARTIAL_CALL_DATE = "RedemptionAtPartialCallDate"
    REDEMPTION_AT_PARTIAL_PUT_DATE = "RedemptionAtPartialPutDate"
    REDEMPTION_AT_PERPETUITY = "RedemptionAtPerpetuity"
    REDEMPTION_AT_PREMIUM_DATE = "RedemptionAtPremiumDate"
    REDEMPTION_AT_PUT_DATE = "RedemptionAtPutDate"
    REDEMPTION_AT_SINK_DATE = "RedemptionAtSinkDate"
    REDEMPTION_AT_WORST_DATE = "RedemptionAtWorstDate"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(RedemptionDateType, _REDEMPTION_DATE_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(
            _REDEMPTION_DATE_TYPE_VALUES_IN_LOWER_BY_REDEMPTION_DATE_TYPE, some
        )


_REDEMPTION_DATE_TYPE_VALUES = tuple(t.value for t in RedemptionDateType)
_REDEMPTION_DATE_TYPE_VALUES_IN_LOWER_BY_REDEMPTION_DATE_TYPE = {
    name.lower(): item for name, item in RedemptionDateType.__members__.items()
}



from typing import List, TYPE_CHECKING

from numpy import iterable

from ...._content_provider import ContentProviderLayer
from ...._content_type import ContentType
from ...._types import ExtendedParams
from ....._tools import create_repr


if TYPE_CHECKING:
    from .. import zc_curve_definition


class Definition(ContentProviderLayer):
    """

    Parameters
    ----------
    universe : list of zc_curve_definition.Definition
        See detail class zc_curve_definition.Definition.
    extended_params : dict, optional
        If necessary other parameters.

    Methods
    -------
    get_data(session=session, on_response=on_response, **kwargs)
        Returns a response to the data platform
    get_data_async(session=None, on_response=None, **kwargs)
        Returns a response asynchronously to the data platform

    Examples
    --------
    >>> from refinitiv.data.content.ipa.curves import zc_curve_definition
    >>> from refinitiv.data.content.ipa.curves import zc_curve_definitions
    >>> definition1 = zc_curve_definition.Definition(source="Refinitiv")
    >>> definition2 = zc_curve_definition.Definition(source="Peugeot")
    >>> definitions = zc_curve_definitions.Definition(
    ...     universe=[definition1, definition2]
    ...)
    >>> response = definitions.get_data()

    Using get_data_async
     >>> import asyncio
     >>> task = definitions.get_data_async()
     >>> response = asyncio.run(task)
    """

    def __init__(
        self,
        universe: List["zc_curve_definition.Definition"],
        extended_params: ExtendedParams = None,
    ):

        if not iterable(universe):
            universe = [universe]

        super().__init__(
            content_type=ContentType.ZC_CURVE_DEFINITIONS,
            universe=universe,
            extended_params=extended_params,
        )

    def __repr__(self):
        return create_repr(self)



# coding: utf8

from .._forward_curve_types import OptFloat
from ..._object_definition import ObjectDefinition


class ConvexityAdjustment(ObjectDefinition):
    """
    Parameters
    ----------
    mean_reversion_percent : float, optional
        Reversion speed rate, expressed in percents, used to calculate the convexity
        adjustment
    volatility_percent : float, optional
        Reversion flat volatility, expressed in percents, used to calculate the
        convexity adjustment
    """

    def __init__(
        self,
        mean_reversion_percent: OptFloat = None,
        volatility_percent: OptFloat = None,
    ) -> None:
        super().__init__()
        self.mean_reversion_percent = mean_reversion_percent
        self.volatility_percent = volatility_percent

    @property
    def mean_reversion_percent(self):
        """
        Reversion speed rate, expressed in percents, used to calculate the convexity
        adjustment
        :return: float
        """
        return self._get_parameter("meanReversionPercent")

    @mean_reversion_percent.setter
    def mean_reversion_percent(self, value):
        self._set_parameter("meanReversionPercent", value)

    @property
    def volatility_percent(self):
        """
        Reversion flat volatility, expressed in percents, used to calculate the
        convexity adjustment
        :return: float
        """
        return self._get_parameter("volatilityPercent")

    @volatility_percent.setter
    def volatility_percent(self, value):
        self._set_parameter("volatilityPercent", value)



# coding: utf8

import itertools
from concurrent.futures import ThreadPoolExecutor, wait
from typing import (
    Union,
    Callable,
    Iterable,
    TYPE_CHECKING,
    Tuple,
    ValuesView,
    KeysView,
    ItemsView,
    Dict,
)

import pandas as pd

from ._streamingprice import StreamingPrice
from .._types import Strings
from ..._core.session._default_session_manager import get_valid_session
from ..._tools import universe_arg_parser, fields_arg_parser, cached_property
from ...delivery._stream import (
    OMMStreamListener,
    StreamEvent,
    StreamStateManager,
    StreamStateEvent,
)
from ...delivery._stream.base_stream import StreamOpenWithUpdatesMixin
from ..._errors import ItemWasNotRequested

if TYPE_CHECKING:
    from ..._core.session._session import Session


def validate(name, input_values, requested_values):
    input_values = set(input_values)
    requested_values = set(requested_values)
    not_requested = input_values - requested_values
    has_not_requested = bool(not_requested)

    if has_not_requested:
        raise ItemWasNotRequested(name, not_requested, requested_values)


def get_available_fields(value_by_fields_by_inst_names):
    available_fields = set()
    for value_by_fields in value_by_fields_by_inst_names.values():
        fields = value_by_fields.keys()
        available_fields.update(fields)
    return available_fields


class PricingStream(StreamOpenWithUpdatesMixin):
    def __init__(self, stream: StreamingPrice):
        self._stream: StreamingPrice = stream

    @property
    def id(self) -> int:
        return self._stream.id

    @property
    def code(self):
        return self._stream.code

    @property
    def message(self):
        return self._stream.message

    @property
    def name(self):
        return self._stream.name

    @property
    def service(self):
        return self._stream.service

    @property
    def fields(self):
        return self._stream.fields

    @property
    def status(self):
        return self._stream.status

    @property
    def is_ok(self):
        return self._stream.is_ok

    def __enter__(self):
        return self._stream.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self._stream.__exit__(exc_type, exc_val, exc_tb)

    def pause(self):
        return self._stream.pause()

    def resume(self):
        return self._stream.resume()

    def get_field_value(self, field):
        return self._stream.get_field_value(field)

    def get_fields(self, fields=None) -> dict:
        return self._stream.get_fields(fields)

    def keys(self):
        return self._stream.keys()

    def values(self):
        return self._stream.values()

    def items(self):
        return self._stream.items()

    def __iter__(self):
        return self._stream.__iter__()

    def __getitem__(self, field):
        return self._stream.__getitem__(field)

    def __len__(self):
        return self._stream.__len__()

    def __repr__(self):
        return self._stream.__repr__()

    def __str__(self):
        return self._stream.__str__()


_id_iterator = itertools.count()


class StreamingPrices(StreamStateManager, OMMStreamListener["StreamingPrices"]):
    def __init__(
        self,
        universe: Union[str, Iterable[str]],
        session: "Session" = None,
        fields: Union[str, list] = None,
        service: str = None,
        on_refresh: Callable = None,
        on_status: Callable = None,
        on_update: Callable = None,
        on_complete: Callable = None,
        extended_params: dict = None,
    ) -> None:
        session = get_valid_session(session)

        StreamStateManager.__init__(self, logger=session.logger())
        OMMStreamListener.__init__(
            self,
            logger=session.logger(),
            on_refresh=on_refresh,
            on_status=on_status,
            on_update=on_update,
            on_complete=on_complete,
        )

        self.universe: Strings = universe_arg_parser.get_list(universe)
        self._session = session
        self.fields: Strings = fields_arg_parser.get_list(fields or [])
        self._service = service
        self._extended_params = extended_params
        self._id = next(_id_iterator)
        self._classname: str = (
            f"[{self.__class__.__name__} id={self._id} universe={universe}]"
        )
        self._completed = set()

    @cached_property
    def _streaming_prices_by_name(self) -> Dict[str, StreamingPrice]:
        retval = {}
        for name in self.universe:
            streaming_price = StreamingPrice(
                session=self._session,
                name=name,
                fields=self.fields,
                service=self._service,
                on_refresh=self._on_stream_refresh,
                on_status=self._on_stream_status,
                on_update=self._on_stream_update,
                on_complete=self._on_stream_complete,
                on_error=self._on_stream_error,
                extended_params=self._extended_params,
                parent_id=self._id,
            )

            streaming_price.on(StreamStateEvent.CLOSED, self._on_stream_close)
            retval[name] = streaming_price
        return retval

    def keys(self) -> KeysView[str]:
        return self._streaming_prices_by_name.keys()

    def values(self) -> ValuesView[StreamingPrice]:
        return self._streaming_prices_by_name.values()

    def items(self) -> ItemsView[str, StreamingPrice]:
        return self._streaming_prices_by_name.items()

    def __iter__(self):
        return StreamingPricesIterator(self)

    def __getitem__(self, name) -> PricingStream:
        streaming_price = self._streaming_prices_by_name.get(name, {})

        if streaming_price == {}:
            self._warning(f"'{name}' not in {self._classname} universe")
            return streaming_price

        if hasattr(streaming_price, "_wrapper"):
            wrapper = streaming_price._wrapper
        else:
            wrapper = PricingStream(streaming_price)
            streaming_price._wrapper = wrapper

        return wrapper

    def __len__(self):
        return len(self._streaming_prices_by_name)

    def get_snapshot(
        self, universe: Strings = None, fields: Strings = None, convert: bool = True
    ) -> pd.DataFrame:
        """
        Returns a Dataframe filled with snapshot values
        for a list of instrument names and a list of fields.

        Parameters
        ----------
        universe: list of strings
            List of instruments to request snapshot data on.

        fields: list of strings
            List of fields to request.

        convert: boolean
            If True, force numeric conversion for all values.

        Returns
        -------
            pandas.DataFrame

            pandas.DataFrame content:
                - columns : instrument and field names
                - rows : instrument name and field values

        Raises
        ------
            Exception
                If request fails or if server returns an error

            ValueError
                If a parameter type or value is wrong

        """

        if universe:
            universe = universe_arg_parser.get_list(universe)
            try:
                validate("Instrument", universe, self.universe)
            except ItemWasNotRequested as e:
                self._error(e)
                return pd.DataFrame()
        else:
            universe = self.universe

        value_by_fields_by_inst_names: dict = {}
        for name in universe:
            streaming_price: StreamingPrice = self._streaming_prices_by_name.get(name)

            values_by_fields = streaming_price.get_fields(fields)
            if values_by_fields:
                value_by_fields_by_inst_names[name] = values_by_fields

        if fields:
            try:
                self.fields and validate("Field", fields, self.fields)
            except ItemWasNotRequested as e:
                self._error(e)
                return pd.DataFrame()
        else:
            fields = get_available_fields(value_by_fields_by_inst_names)

        values_by_field = {}
        for field in fields:
            values_by_field[field] = [
                value_by_fields_by_inst_names.get(name, {}).get(field, None)
                for name in universe
            ]

        price_df = pd.DataFrame(values_by_field, columns=fields)

        if convert and not price_df.empty:
            price_df = price_df.convert_dtypes()

        universe_df = pd.DataFrame(universe, columns=["Instrument"])
        return pd.concat([universe_df, price_df], axis=1)

    def _do_pause(self) -> None:
        for streaming_price in self.values():
            streaming_price.pause()

    def _do_resume(self) -> None:
        for streaming_price in self.values():
            streaming_price.resume()

    def _do_open(self, *args, with_updates=True) -> None:
        self._debug(f"{self._classname} open streaming on {self.universe}")

        if not self.values():
            raise ValueError("No instrument to subscribe")

        self._completed.clear()

        with ThreadPoolExecutor(
            thread_name_prefix="OpenStreamingPrices-Thread"
        ) as executor:
            futures = [
                executor.submit(stream.open, with_updates=with_updates)
                for stream in self.values()
            ]
            wait(futures)
            for fut in futures:
                exception = fut.exception()
                if exception:
                    raise exception

        self._debug(f"{self._classname} streaming on {self.universe} is open")

    def _do_close(self, *args, **kwargs) -> None:
        self._debug(f"{self._classname} close streaming on {str(self.universe)}")

        with ThreadPoolExecutor(
            thread_name_prefix="CloseStreamingPrices-Thread"
        ) as executor:
            futures = [executor.submit(stream.close) for stream in self.values()]
            wait(futures)
            for fut in futures:
                exception = fut.exception()
                if exception:
                    raise exception

    def _on_stream_close(self, streaming_price: StreamingPrice) -> None:
        self._debug(f"{self._classname} streaming closed name={streaming_price.name}")
        self.close()

    def _do_on_stream_refresh(self, stream: StreamingPrice, *args) -> Tuple:
        return self._do_on_stream(stream, *args)

    def _on_stream_status(self, stream: StreamingPrice, *args) -> None:
        self._debug(f"{self._classname} on_status {args}")

        status = args[0]
        self._emitter.emit(StreamEvent.STATUS, self, stream.name, status)

        if stream.is_closed:
            self.dispatch_complete(stream)

    def _do_on_stream_update(self, stream: StreamingPrice, *args) -> Tuple:
        return self._do_on_stream(stream, *args)

    def _do_on_stream(self, stream: StreamingPrice, *args) -> Tuple:
        fields = args[0]
        return stream.name, fields

    def _on_stream_complete(self, stream: StreamingPrice, *args) -> None:
        self._debug(f"{self._classname} on_complete {args}")

        name = stream.name
        if name in self._completed:
            return

        self._completed.update([name])
        if self._completed == set(self.universe):
            # received complete event from all streams all, emit global complete
            self._emitter.emit(StreamEvent.COMPLETE, self)

    def _do_on_stream_error(self, stream: StreamingPrice, *args) -> Tuple:
        error = args[0]
        return stream.name, error


class StreamingPricesIterator:
    def __init__(self, streaming_prices: StreamingPrices):
        self._streaming_prices = streaming_prices
        self._index = 0

    def __next__(self):
        if self._index < len(self._streaming_prices.universe):
            result = self._streaming_prices[
                self._streaming_prices.universe[self._index]
            ]
            self._index += 1
            return result
        raise StopIteration()



# coding: utf8
# contract_gen 2020-06-16 10:26:10.708191

__all__ = ["FxVolatilityModel"]

from enum import Enum, unique

from ._common_tools import _convert_to_str, _normalize


@unique
class FxVolatilityModel(Enum):
    CUBIC_SPLINE = "CubicSpline"
    SABR = "SABR"
    SVI = "SVI"
    TWIN_LOGNORMAL = "TwinLognormal"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(FxVolatilityModel, _FX_VOLATILITY_MODEL_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(
            _FX_VOLATILITY_MODEL_VALUES_IN_LOWER_BY_FX_VOLATILITY_MODEL, some
        )


_FX_VOLATILITY_MODEL_VALUES = tuple(t.value for t in FxVolatilityModel)
_FX_VOLATILITY_MODEL_VALUES_IN_LOWER_BY_FX_VOLATILITY_MODEL = {
    name.lower(): item for name, item in FxVolatilityModel.__members__.items()
}



# coding: utf8

__all__ = ["LegDefinition"]

from ._enums import IndexSpreadCompoundingMethod
from ._enums import AdjustInterestToPaymentDate
from ._enums import DayCountBasis
from ._enums import BusinessDayConvention
from ._enums import Direction
from ._enums import Frequency
from ._enums import IndexCompoundingMethod
from ._enums import IndexResetType
from ._enums import InterestType
from ._enums import NotionalExchange
from ._enums import DateRollingConvention
from ._enums import StubRule
from ..._object_definition import ObjectDefinition
from ._models import AmortizationItem


class LegDefinition(ObjectDefinition):
    def __init__(
        self,
        *,
        leg_tag=None,
        direction=None,
        interest_type=None,
        notional_ccy=None,
        notional_amount=None,
        fixed_rate_percent=None,
        index_name=None,
        index_tenor=None,
        index_fixing_ric=None,
        spread_bp=None,
        interest_payment_frequency=None,
        interest_calculation_method=None,
        accrued_calculation_method=None,
        payment_business_day_convention=None,
        payment_roll_convention=None,
        index_reset_frequency=None,
        index_reset_type=None,
        index_fixing_lag=None,
        first_regular_payment_date=None,
        last_regular_payment_date=None,
        amortization_schedule=None,
        payment_business_days=None,
        notional_exchange=None,
        adjust_interest_to_payment_date=None,
        index_compounding_method=None,
        interest_payment_delay=None,
        stub_rule=None,
        index_spread_compounding_method=None,
        cms_template=None,
        floor_strike_percent=None,
    ):
        super().__init__()
        self.leg_tag = leg_tag
        self.direction = direction
        self.interest_type = interest_type
        self.notional_ccy = notional_ccy
        self.notional_amount = notional_amount
        self.fixed_rate_percent = fixed_rate_percent
        self.index_name = index_name
        self.index_tenor = index_tenor
        self.spread_bp = spread_bp
        self.interest_payment_frequency = interest_payment_frequency
        self.interest_calculation_method = interest_calculation_method
        self.accrued_calculation_method = accrued_calculation_method
        self.payment_business_day_convention = payment_business_day_convention
        self.payment_roll_convention = payment_roll_convention
        self.index_reset_frequency = index_reset_frequency
        self.index_reset_type = index_reset_type
        self.index_fixing_lag = index_fixing_lag
        self.first_regular_payment_date = first_regular_payment_date
        self.last_regular_payment_date = last_regular_payment_date
        self.amortization_schedule = amortization_schedule
        self.payment_business_days = payment_business_days
        self.notional_exchange = notional_exchange
        self.adjust_interest_to_payment_date = adjust_interest_to_payment_date
        self.index_compounding_method = index_compounding_method
        self.interest_payment_delay = interest_payment_delay
        self.stub_rule = stub_rule
        self.index_fixing_ric = index_fixing_ric
        self.index_spread_compounding_method = index_spread_compounding_method
        self.cms_template = cms_template
        self.floor_strike_percent = floor_strike_percent

    @property
    def accrued_calculation_method(self):
        """
        The Day Count Basis method used to calculate the accrued interest payments.
        Optional. By default, the same value than InterestCalculationMethod is used.
        :return: enum DayCountBasis
        """
        return self._get_enum_parameter(DayCountBasis, "accruedCalculationMethod")

    @accrued_calculation_method.setter
    def accrued_calculation_method(self, value):
        self._set_enum_parameter(DayCountBasis, "accruedCalculationMethod", value)

    @property
    def adjust_interest_to_payment_date(self):
        """
        A flag that indicates if the coupon dates are adjusted to the payment dates.
        Optional. By default 'false' is used.
        :return: enum AdjustInterestToPaymentDate
        """
        return self._get_enum_parameter(
            AdjustInterestToPaymentDate, "adjustInterestToPaymentDate"
        )

    @adjust_interest_to_payment_date.setter
    def adjust_interest_to_payment_date(self, value):
        self._set_enum_parameter(
            AdjustInterestToPaymentDate, "adjustInterestToPaymentDate", value
        )

    @property
    def amortization_schedule(self):
        """
        Definition of amortizations
        :return: list AmortizationItem
        """
        return self._get_list_parameter(AmortizationItem, "amortizationSchedule")

    @amortization_schedule.setter
    def amortization_schedule(self, value):
        self._set_list_parameter(AmortizationItem, "amortizationSchedule", value)

    @property
    def direction(self):
        """
        The direction of the leg. the possible values are:

         'Paid' (the cash flows of the leg are paid to the counterparty),

         'Received' (the cash flows of the leg are received from the counterparty).

        Optional for a single leg instrument (like a bond), in that case default value is Received. It is mandatory for a
        multi-instrument leg instrument (like Swap or CDS leg).
        :return: enum Direction
        """
        return self._get_enum_parameter(Direction, "direction")

    @direction.setter
    def direction(self, value):
        self._set_enum_parameter(Direction, "direction", value)

    @property
    def index_compounding_method(self):
        """
        A flag that defines how the coupon rate is calculated from the reset floating rates when the reset frequency is higher than
        the interest payment frequency (e.g. daily index reset with quarterly interest payment).
        The possible values are:
         - Compounded (uses the compounded average rate from multiple fixings),
         - Average (uses the arithmetic average rate from multiple fixings),
         - Constant (uses the last published rate among multiple fixings),
         - AdjustedCompounded (uses Chinese 7-day repo fixing),
         - MexicanCompounded (uses Mexican Bremse fixing).
        Optional. By default 'Constant' is used.
        :return: enum IndexCompoundingMethod
        """
        return self._get_enum_parameter(
            IndexCompoundingMethod, "indexCompoundingMethod"
        )

    @index_compounding_method.setter
    def index_compounding_method(self, value):
        self._set_enum_parameter(
            IndexCompoundingMethod, "indexCompoundingMethod", value
        )

    @property
    def index_reset_frequency(self):
        """
        The reset frequency in case the leg Type is Float.
        Optional. By default the IndexTenor is used.
        :return: enum Frequency
        """
        return self._get_enum_parameter(Frequency, "indexResetFrequency")

    @index_reset_frequency.setter
    def index_reset_frequency(self, value):
        self._set_enum_parameter(Frequency, "indexResetFrequency", value)

    @property
    def index_reset_type(self):
        """
        A flag that indicates if the floating rate index is reset before the coupon period starts or at the end of the coupon period.
        The possible values are:

         - InAdvance (resets the index before the start of the interest period),

         - InArrears (resets the index at the end of the interest period).

        Optional. By default 'InAdvance' is used.
        :return: enum IndexResetType
        """
        return self._get_enum_parameter(IndexResetType, "indexResetType")

    @index_reset_type.setter
    def index_reset_type(self, value):
        self._set_enum_parameter(IndexResetType, "indexResetType", value)

    @property
    def interest_calculation_method(self):
        """
        The Day Count Basis method used to calculate the coupon interest payments.
        Mandatory.
        :return: enum DayCountBasis
        """
        return self._get_enum_parameter(DayCountBasis, "interestCalculationMethod")

    @interest_calculation_method.setter
    def interest_calculation_method(self, value):
        self._set_enum_parameter(DayCountBasis, "interestCalculationMethod", value)

    @property
    def interest_payment_frequency(self):
        """
        The frequency of the interest payments.
        Optional if an instrument code/style have been defined : in that case, value comes from reference data. Otherwise,
        it is mandatory.
        :return: enum Frequency
        """
        return self._get_enum_parameter(Frequency, "interestPaymentFrequency")

    @interest_payment_frequency.setter
    def interest_payment_frequency(self, value):
        self._set_enum_parameter(Frequency, "interestPaymentFrequency", value)

    @property
    def interest_type(self):
        """
        A flag that indicates whether the leg is fixed or float. Possible values are:
        - 'Fixed' (the leg has a fixed coupon),
        - 'Float' (the leg has a floating rate index).
        Mandatory.
        :return: enum InterestType
        """
        return self._get_enum_parameter(InterestType, "interestType")

    @interest_type.setter
    def interest_type(self, value):
        self._set_enum_parameter(InterestType, "interestType", value)

    @property
    def notional_exchange(self):
        """
        A flag that defines whether and when notional payments occurs.
        The possible values are:

         - None (means that the notional is not included in the cash flow schedule),

         - Start (means that the counterparties exchange the notional on the swap start date, before the first interest payment),

         - End (means that the counterparties exchange the notional on the swap maturity date, in addition to the last interest
         payment),

         - Both (combines the payments of Start Only and End Only),

         - EndAdjustment.
        :return: enum NotionalExchange
        """
        return self._get_enum_parameter(NotionalExchange, "notionalExchange")

    @notional_exchange.setter
    def notional_exchange(self, value):
        self._set_enum_parameter(NotionalExchange, "notionalExchange", value)

    @property
    def payment_business_day_convention(self):
        """
        The method to adjust dates to a working day.
        The possible values are:
         - ModifiedFollowing (adjusts dates according to the Modified Following convention - next business day unless is it goes
         into the next month, preceeding is used in that  case),
         - NextBusinessDay (adjusts dates according to the Following convention - Next Business Day),
         - PreviousBusinessDay (adjusts dates  according to the Preceeding convention - Previous Business Day),
         - NoMoving (does not adjust dates),
         - BbswModifiedFollowing (adjusts dates  according to the BBSW Modified Following convention).
        Optional. In case an instrument code/style has been defined, value comes from bond reference data. Otherwise
        'ModifiedFollowing' is used.
        :return: enum BusinessDayConvention
        """
        return self._get_enum_parameter(
            BusinessDayConvention, "paymentBusinessDayConvention"
        )

    @payment_business_day_convention.setter
    def payment_business_day_convention(self, value):
        self._set_enum_parameter(
            BusinessDayConvention, "paymentBusinessDayConvention", value
        )

    @property
    def payment_roll_convention(self):
        """
        The method to adjust payment dates whn they fall at the end of the month (28th of February, 30th, 31st).
        The possible values are:

         - Last (For setting the calculated date to the last working day),

         - Same (For setting the calculated date to the same day . In this latter case, the date may be moved according to the date
         moving convention if it is a non-working day),

         - Last28 (For setting the calculated date to the last working day. 28FEB being always considered as the last working day),

         - Same28 (For setting the calculated date to the same day .28FEB being always considered as the last working day).

        Optional. By default 'SameDay' is used.
        :return: enum DateRollingConvention
        """
        return self._get_enum_parameter(DateRollingConvention, "paymentRollConvention")

    @payment_roll_convention.setter
    def payment_roll_convention(self, value):
        self._set_enum_parameter(DateRollingConvention, "paymentRollConvention", value)

    @property
    def stub_rule(self):
        """
        The rule that defines whether coupon roll dates are aligned on the  maturity or the issue date.
        The possible values are:
         - ShortFirstProRata (to create a short period between the start date and the first coupon date, and pay a smaller amount of
         interest for the short period.All coupon dates are calculated backward from the maturity date),
         - ShortFirstFull (to create a short period between the start date and the first coupon date, and pay a regular coupon on
         the first coupon date. All coupon dates are calculated backward from the maturity date),
         - LongFirstFull (to create a long period between the start date and the second coupon date, and pay a regular coupon on the
         second coupon date. All coupon dates are calculated backward from the maturity date),
         - ShortLastProRata (to create a short period between the last payment date and maturity, and pay a smaller amount of
         interest for the short period. All coupon dates are calculated forward from the start date).
        This property may also be used in conjunction with firstRegularPaymentDate and lastRegularPaymentDate; in that case the
        following values can be defined:
         - Issue (all dates are aligned on the issue date),
         - Maturity (all dates are aligned on the maturity date).
        Optional. By default 'Maturity' is used.
        :return: enum StubRule
        """
        return self._get_enum_parameter(StubRule, "stubRule")

    @stub_rule.setter
    def stub_rule(self, value):
        self._set_enum_parameter(StubRule, "stubRule", value)

    @property
    def first_regular_payment_date(self):
        """
        The first regular coupon payment date for leg with an odd first coupon.
        Optional.
        :return: str
        """
        return self._get_parameter("firstRegularPaymentDate")

    @first_regular_payment_date.setter
    def first_regular_payment_date(self, value):
        self._set_parameter("firstRegularPaymentDate", value)

    @property
    def fixed_rate_percent(self):
        """
        The fixed coupon rate in percentage.
        It is mandatory in case of a single leg instrument. Otherwise, in case of multi leg instrument, it can be computed as the
        Par rate.
        :return: float
        """
        return self._get_parameter("fixedRatePercent")

    @fixed_rate_percent.setter
    def fixed_rate_percent(self, value):
        self._set_parameter("fixedRatePercent", value)

    @property
    def index_fixing_lag(self):
        """
        Defines the number of working days between the fixing date and the start of the coupon period ('InAdvance') or the end of
        the coupon period ('InArrears').
        Optional. By default 0 is used.
        :return: int
        """
        return self._get_parameter("indexFixingLag")

    @index_fixing_lag.setter
    def index_fixing_lag(self, value):
        self._set_parameter("indexFixingLag", value)

    @property
    def index_fixing_ric(self):
        """
        The RIC that carries the fixing value. This value overrides the RIC associated by default with the IndexName and IndexTenor.
        Optional.
        :return: str
        """
        return self._get_parameter("indexFixingRic")

    @index_fixing_ric.setter
    def index_fixing_ric(self, value):
        self._set_parameter("indexFixingRic", value)

    @property
    def index_name(self):
        """
        The name of the floating rate index.
        Mandatory when the leg is float.
        :return: str
        """
        return self._get_parameter("indexName")

    @index_name.setter
    def index_name(self, value):
        self._set_parameter("indexName", value)

    @property
    def index_tenor(self):
        """
        The period code that represents the maturity of the floating rate index.
        Mandatory when the leg is float.
        :return: str
        """
        return self._get_parameter("indexTenor")

    @index_tenor.setter
    def index_tenor(self, value):
        self._set_parameter("indexTenor", value)

    @property
    def interest_payment_delay(self):
        """
        The number of working days between the end of coupon period and the actual interest payment date.
        Optional. By default no delay (0) is applied.
        :return: int
        """
        return self._get_parameter("interestPaymentDelay")

    @interest_payment_delay.setter
    def interest_payment_delay(self, value):
        self._set_parameter("interestPaymentDelay", value)

    @property
    def last_regular_payment_date(self):
        """
        The last regular coupon payment date for leg with an odd last coupon.
        Optional.
        :return: str
        """
        return self._get_parameter("lastRegularPaymentDate")

    @last_regular_payment_date.setter
    def last_regular_payment_date(self, value):
        self._set_parameter("lastRegularPaymentDate", value)

    @property
    def leg_tag(self):
        """
        A user provided string to identify the leg  that will also be part of the response.
        Optional.
        :return: str
        """
        return self._get_parameter("legTag")

    @leg_tag.setter
    def leg_tag(self, value):
        self._set_parameter("legTag", value)

    @property
    def notional_amount(self):
        """
        The notional amount of the leg at the period start date.
        Optional. By default 1,000,000 is used.
        :return: float
        """
        return self._get_parameter("notionalAmount")

    @notional_amount.setter
    def notional_amount(self, value):
        self._set_parameter("notionalAmount", value)

    @property
    def notional_ccy(self):
        """
        The ISO code of the notional currency.
        Mandatory if instrument code or instrument style has not been defined. In case an instrument code/style has been defined,
        value may comes from the reference data.
        :return: str
        """
        return self._get_parameter("notionalCcy")

    @notional_ccy.setter
    def notional_ccy(self, value):
        self._set_parameter("notionalCcy", value)

    @property
    def payment_business_days(self):
        """
        A list of coma-separated calendar codes to adjust dates (e.g. 'EMU' or 'USA').
        Optional. By default the calendar associated to NotionalCcy is used.
        :return: str
        """
        return self._get_parameter("paymentBusinessDays")

    @payment_business_days.setter
    def payment_business_days(self, value):
        self._set_parameter("paymentBusinessDays", value)

    @property
    def spread_bp(self):
        """
        The spread in basis point that is added to the floating rate index index value.
        Optional. By default 0 is used.
        :return: float
        """
        return self._get_parameter("spreadBp")

    @spread_bp.setter
    def spread_bp(self, value):
        self._set_parameter("spreadBp", value)

    @property
    def index_spread_compounding_method(self):
        """
        A flag that defines how the computed spread of the floating leg is applied to the multiple fixing rates when compounded.
        Optional. By default, when IndexCompoundingMethod=Compounded,  'Compounded' is used.
        :return: enum IndexSpreadCompoundingMethod
        """
        return self._get_enum_parameter(
            IndexSpreadCompoundingMethod, "indexSpreadCompoundingMethod"
        )

    @index_spread_compounding_method.setter
    def index_spread_compounding_method(self, value):
        self._set_enum_parameter(
            IndexSpreadCompoundingMethod, "indexSpreadCompoundingMethod", value
        )

    @property
    def cms_template(self):
        """
        A reference to a common swap contract that represents the underlying swap in case of a Constant Maturity Swap contract (CMS).
        Mandatory for CMS contract.
        :return: str
        """
        return self._get_parameter("cmsTemplate")

    @cms_template.setter
    def cms_template(self, value):
        self._set_parameter("cmsTemplate", value)

    @property
    def floor_strike_percent(self):
        """
        :return: float
        """
        return self._get_parameter("floorStrikePercent")

    @floor_strike_percent.setter
    def floor_strike_percent(self, value):
        self._set_parameter("floorStrikePercent", value)



__all__ = (
    "forward_curve",
    "forward_curves",
    "zc_curve",
    "zc_curve_definition",
    "zc_curve_definitions",
    "zc_curves",
)

from . import forward_curve
from . import forward_curves
from . import zc_curve
from . import zc_curve_definition
from . import zc_curve_definitions
from . import zc_curves



# coding: utf8

from typing import Optional, List

from ._base_definition import BaseDefinition, FCBaseDefinition
from ._instrument_pricing_parameters import (
    InstrumentPricingParameters as PricingParameters,
)
from ...._tools import create_repr


class Definition(FCBaseDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    universe : list
        Array of Financial Contract definitions.
    fields: list of str, optional
        Array of requested fields. each requested field will represent
        a column of the tabular response. By default all relevant fields
        are returned.
    pricing_parameters : PricingParameters, optional
        Pricing parameters that are specific to the financial contracts
        defined in universe.
    extended_params : dict, optional
        If necessary other parameters

    Methods
    -------
    get_data(session=session, on_response=on_response, async_mode=None)
        Returns a response to the data platform
    get_data_async(session=session, on_response=on_response, async_mode=None)
        Returns a response to the async data platform

    Examples
    --------
     >>> import refinitiv.data.content.ipa.financial_contracts as rdf
     >>> option_definition = rdf.option.Definition(
     ...     instrument_code="FCHI560000L1.p",
     ...     underlying_type=rdf.option.UnderlyingType.ETI,
     ...     fields=[
     ...         "MarketValueInDealCcy",
     ...         "DeltaPercent",
     ...         "GammaPercent",
     ...         "RhoPercent",
     ...         "ThetaPercent",
     ...         "VegaPercent",
     ...         "ErrorCode",
     ...         "ErrorMessage",
     ...     ],
     ... )
     >>> bond_definition = rdf.bond.Definition(
     ...     issue_date="2002-02-28",
     ...     end_date="2032-02-28",
     ...     notional_ccy="USD",
     ...     interest_payment_frequency="Annual",
     ...     fixed_rate_percent=7,
     ...     interest_calculation_method=rdf.bond.DayCountBasis.DCB_ACTUAL_ACTUAL
     ... )
     >>> definition = rdf.Definition(
     ...    [
     ...        bond_definition,
     ...        option_definition
     ...    ]
     ... )
     >>> response = definition.get_data()
    """

    def __init__(
        self,
        universe: List[BaseDefinition],
        fields: Optional[List[str]] = None,
        pricing_parameters: Optional[PricingParameters] = None,
        extended_params: Optional[dict] = None,
    ) -> None:
        super().__init__(
            universe=universe,
            fields=fields,
            pricing_parameters=pricing_parameters,
            extended_params=extended_params,
        )

    def __repr__(self):
        return create_repr(self)



from concurrent.futures import Future, wait
from typing import TYPE_CHECKING, Dict, Any

from ._chain_record import create_chain_record, can_create_chain_record
from ..._content_type import ContentType
from ..._types import ExtendedParams, OptStr
from ....delivery._stream import StreamStateManager, OMMStreamListener
from ....delivery._stream._stream_factory import create_omm_stream

if TYPE_CHECKING:
    from ._chain_record import ChainRecord
    from ...._core.session import Session
    from ....delivery._stream import _OMMStream


class ChainRecords(StreamStateManager, OMMStreamListener["ChainRecords"]):
    def __init__(
        self,
        session: "Session",
        service: OptStr = None,
        extended_params: ExtendedParams = None,
    ):
        StreamStateManager.__init__(self, logger=session.logger())
        OMMStreamListener.__init__(self, logger=session.logger())
        self.display_template = None
        self.service: OptStr = service
        self.session: "Session" = session
        self.extended_params = extended_params
        self.records_by_name: Dict[str, "ChainRecord"] = {}
        self.refreshing_by_name: Dict[str, "Future"] = {}
        self.streams_by_name: Dict[str, "_OMMStream"] = {}

    def add(self, name: str):
        stream = create_omm_stream(
            ContentType.STREAMING_CHAINS,
            session=self.session,
            name=name,
            domain="MarketPrice",
            service=self.service,
            fields=[],
            extended_params=self.extended_params,
        )
        stream.on_refresh(self._on_stream_refresh)
        stream.on_status(self._on_stream_status)
        stream.on_update(self._on_stream_update)
        stream.on_complete(self._on_stream_complete)
        stream.on_error(self._on_stream_error)

        self.streams_by_name[name] = stream
        self.refreshing_by_name[name] = Future()
        self._debug(f"{self._classname} added stream for name: {name}")
        return stream

    def has(self, name: str) -> bool:
        return name in self.records_by_name

    def has_stream(self, name: str) -> bool:
        return name in self.streams_by_name

    def not_has_stream(self, name: str) -> bool:
        return not self.has_stream(name)

    def is_stream_close(self, name: str) -> bool:
        return self.streams_by_name[name].is_close

    def get_display_name(self, name: str) -> str:
        chain_record = self.records_by_name[name]
        return chain_record.display_name

    def _do_open(self, with_updates: bool):
        for stream in self.streams_by_name.values():
            self._debug(f"{self._classname} opens a stream with name {stream.name} [o]")
            stream.open(with_updates=with_updates)

    def _do_close(self):
        for refreshing in self.refreshing_by_name.values():
            if not refreshing.done():
                refreshing.set_result(1)

        for stream in self.streams_by_name.values():
            self._debug(
                f"{self._classname} closes a stream with name {stream.name} [c]"
            )
            stream.close()

    def open_stream(self, name: str, with_updates: bool):
        stream = self.streams_by_name[name]
        self._debug(f"{self._classname} opens a stream with name {stream.name} [c]")
        stream.open(with_updates=with_updates)

    def wait_refresh(self, name: str):
        wait([self.refreshing_by_name[name]])

    def get_record(self, name: str) -> "ChainRecord":
        record = self.records_by_name[name] if name in self.records_by_name else None
        return record

    def get_stream(self, name: str) -> "_OMMStream":
        return self.streams_by_name[name]

    def _do_on_stream_refresh(self, stream: "_OMMStream", *args) -> Any:
        message = args[0]
        fields = message.get("Fields", [])

        if can_create_chain_record(fields):
            chain_record = create_chain_record(fields)
            self._debug(f"{self._classname} created chain_record={chain_record}")
            self.records_by_name[stream.name] = chain_record

            if not self.display_template:
                self.display_template = chain_record.display_template

            refreshing = self.refreshing_by_name[stream.name]

            if not refreshing.done():
                refreshing.set_result(True)

        else:
            self._error(
                f"StreamingChain :: "
                f"Cannot parse chain {stream.name} because it is an invalid chain."
            )

        return message

    def _do_on_stream_status(self, originator, *args) -> Any:
        message = args[0]
        state = message.get("State", {})
        stream_state = state.get("Stream")
        if stream_state == "Closed":
            self.dispatch_error(originator, message)

        return message

    def _do_on_stream_error(self, originator: "_OMMStream", *args) -> Any:
        return originator.name, args

    def _do_on_stream_update(self, originator: "_OMMStream", *args) -> Any:
        return (originator,) + args



# coding: utf8

__all__ = ["BenchmarkYieldSelectionMode"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class BenchmarkYieldSelectionMode(Enum):
    """
    The benchmark yield selection mode:
        - Interpolate : do an interpolatation on yield curve to compute the reference
          yield.
        - Nearest : use the nearest point to find the reference yield.
    """

    INTERPOLATE = "Interpolate"
    NEAREST = "Nearest"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            BenchmarkYieldSelectionMode, _BENCHMARK_YIELD_SELECTION_MODE_VALUES, some
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _BENCHMARK_YIELD_SELECTION_MODE_VALUES_IN_LOWER_BY_BENCHMARK_YIELD_SELECTION_MODE,
            some,
        )


_BENCHMARK_YIELD_SELECTION_MODE_VALUES = tuple(
    t.value for t in BenchmarkYieldSelectionMode
)
_BENCHMARK_YIELD_SELECTION_MODE_VALUES_IN_LOWER_BY_BENCHMARK_YIELD_SELECTION_MODE = {
    name.lower(): item for name, item in BenchmarkYieldSelectionMode.__members__.items()
}



from ..._ipa_content_provider import IPAContentProviderLayer
from ..._surfaces._swaption_surface_request_item import SwaptionSurfaceRequestItem
from ...._content_type import ContentType


class Definition(IPAContentProviderLayer):
    def __init__(
        self,
        instrument_type=None,
        surface_layout=None,
        surface_parameters=None,
        underlying_definition=None,
        surface_tag=None,
        extended_params=None,
    ):
        request_item = SwaptionSurfaceRequestItem(
            instrument_type=instrument_type,
            surface_layout=surface_layout,
            surface_parameters=surface_parameters,
            underlying_definition=underlying_definition,
            surface_tag=surface_tag,
        )
        super().__init__(
            content_type=ContentType.SURFACES,
            universe=request_item,
            extended_params=extended_params,
        )



# coding: utf8

__all__ = ["ExerciseScheduleType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class ExerciseScheduleType(Enum):
    """
    Possible values:
        - FixedLeg
        - FloatLeg
        - UserDefined
    """

    FIXED_LEG = "FixedLeg"
    FLOAT_LEG = "FloatLeg"
    USER_DEFINED = "UserDefined"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            ExerciseScheduleType, _EXERCISE_SCHEDULE_TYPE_VALUES, some
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _EXERCISE_SCHEDULE_TYPE_VALUES_IN_LOWER_BY_EXERCISE_SCHEDULE_TYPE, some
        )


_EXERCISE_SCHEDULE_TYPE_VALUES = tuple(t.value for t in ExerciseScheduleType)
_EXERCISE_SCHEDULE_TYPE_VALUES_IN_LOWER_BY_EXERCISE_SCHEDULE_TYPE = {
    name.lower(): item for name, item in ExerciseScheduleType.__members__.items()
}



# coding: utf8

__all__ = ["DocClause"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class DocClause(Enum):
    CUM_RESTRUCT14 = "CumRestruct14"
    MODIFIED_RESTRUCT14 = "ModifiedRestruct14"
    MOD_MOD_RESTRUCT14 = "ModModRestruct14"
    EX_RESTRUCT14 = "ExRestruct14"
    CUM_RESTRUCT03 = "CumRestruct03"
    MODIFIED_RESTRUCT03 = "ModifiedRestruct03"
    MOD_MOD_RESTRUCT03 = "ModModRestruct03"
    EX_RESTRUCT03 = "ExRestruct03"
    NONE = "None"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(DocClause, _DOC_CLAUSE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_DOC_CLAUSE_VALUES_IN_LOWER_BY_DOC_CLAUSE, some)


_DOC_CLAUSE_VALUES = tuple(t.value for t in DocClause)
_DOC_CLAUSE_VALUES_IN_LOWER_BY_DOC_CLAUSE = {
    name.lower(): item for name, item in DocClause.__members__.items()
}



import asyncio
import httpx
import mimetypes
import os
import urllib

from uuid import uuid4
from typing import Callable

from ._errors import DownloadFileError
from .._content_provider import (
    ContentProviderLayer,
    UniverseContentValidator,
    ErrorParser,
)
from .._content_type import ContentType
from ...delivery._data._data_provider import (
    DataProvider,
    RequestFactory,
    ResponseFactory,
    BaseResponse,
    Data,
    EndpointData,
    Error,
    _check_response,
)
from ..._tools._common import get_response_reason

# ---------------------------------------------------------------------------
#   Request factory
# ---------------------------------------------------------------------------


class FilingsRequestFactory(RequestFactory):
    def get_url(self, *args, **kwargs):
        url = super().get_url(*args, **kwargs)
        # check for value of the key, must not be None
        if kwargs.get("filename"):
            return url + "/{filename}"
        elif kwargs.get("dcn") or kwargs.get("doc_id") or kwargs.get("filing_id"):
            return url + "/search/{identifier}/{value}"
        else:
            return url

    def get_path_parameters(self, *_, **kwargs) -> dict:
        path_parameters = kwargs.get("path_parameters") or dict()
        # only one of filename, dcn, doc_id and filing_id will be included in the parameter
        if kwargs.get("filename"):
            path_parameters["filename"] = kwargs["filename"]
        else:
            identifiers = {"dcn", "doc_id", "filing_id"}
            for key, value in kwargs.items():
                if key in identifiers and value is not None:
                    path_parameters["identifier"] = key.replace("_", "")
                    path_parameters["value"] = value
        return path_parameters

    def get_header_parameters(self, *args, **kwargs) -> dict:
        # This is the only header available right now. It is hard-coded in their document
        from ...delivery._data._data_provider_factory import get_api_config

        session = args[0]
        api_config = get_api_config(ContentType.FILINGS_RETRIEVAL, session.config)
        return api_config.get_dict("headers")


# ---------------------------------------------------------------------------
#   Data
# ---------------------------------------------------------------------------


class DownloadFileData(EndpointData):
    def __init__(self, file_location: str = None, raw=None, is_success: bool = False):
        super(DownloadFileData, self).__init__(raw)
        self.file_location = file_location
        self.is_success = is_success


class DownloadFilesData:
    def __init__(self, files: list = None):
        self.files = files


# ---------------------------------------------------------------------------
#   Download response
# ---------------------------------------------------------------------------


class DownloadFileResponse:
    def __init__(self, *args, **kwargs):
        self._is_success: bool = kwargs.get("is_success")
        files_data = [kwargs.get("content_data")]
        self.data = DownloadFilesData(files_data)
        self._status = kwargs.get("status")

        err_code = kwargs.get("error_code")
        err_message = kwargs.get("error_message")

        if err_code or err_message:
            if isinstance(err_code, list) and isinstance(err_message, list):
                errors = [Error(code, msg) for code, msg in zip(err_code, err_message)]
            else:
                errors = [Error(err_code, err_message)]
            self.errors = errors
        else:
            self.errors = []
        self._raw_response = kwargs.get("raw_response")


class DownloadAllFileResponse:
    def __init__(self, files: [DownloadFilesData] = None, errors: list = None):
        self.data = DownloadFilesData(files)
        self.errors = errors


# ---------------------------------------------------------------------------
#   File
# ---------------------------------------------------------------------------


class FilingsFile:
    def __init__(
        self, filename: str = None, signed_url: str = None, mimetype: str = None
    ):
        """
        Parameters
        ----------
        filename : str
            Name of the file
        signed_url : str
            Signed URL to download the file
        mimetype : str
            Mime type of the file
        """
        self.filename = filename
        self.signed_url = signed_url
        # Retrieving file using filename doesn't return mimetype of the file
        self.mimetype = mimetype

    @staticmethod
    def _process_response(
        file_location: str = None,
        http_response: httpx.Response = None,
        error: DownloadFileError = None,
    ) -> (DownloadFileResponse, str):
        extension = None
        if http_response is not None:
            data = {
                "is_success": http_response.is_success,
                "status": {
                    "http_status_code": http_response.status_code,
                    "http_reason": get_response_reason(http_response),
                },
                "raw_response": http_response,
            }

            if http_response.is_success:
                extension = mimetypes.guess_extension(
                    http_response.headers.get("content-type")
                )
                data["content_data"] = DownloadFileData(
                    file_location=file_location + extension,
                    raw=http_response.__dict__,
                    is_success=http_response.is_success,
                )
            else:
                data["error_code"] = http_response.status_code
                data["error_message"] = error.message or get_response_reason(
                    http_response
                )
                data["content_data"] = DownloadFileData(
                    file_location=None,
                    raw={
                        "error": {
                            "code": data["error_code"],
                            "description": data["error_message"],
                        }
                    },
                    is_success=False,
                )
        else:
            # Only error before http request will come here
            data = {
                "is_success": False,
                "status": {
                    "http_status_code": None,
                    "http_reason": get_response_reason(http_response),
                },
                "raw_response": http_response,
                "error_code": error.code,
                "error_message": error.message,
            }
            data["content_data"] = DownloadFileData(
                file_location=None,
                raw={
                    "error": {
                        "code": data["error_code"],
                        "description": data["error_message"],
                    }
                },
                is_success=False,
            )

        return (
            DownloadFileResponse(**data),
            extension,
        )

    def _process_exception(
        self, http_response: httpx.Response = None, error: Exception = None
    ) -> DownloadFileResponse:
        exception_response, _ = self._process_response(
            http_response=http_response, error=error
        )
        return exception_response

    def _prepare_and_validate_file_path(self, path):
        file_location = os.path.join(f"{path or os.getcwd()}", f"{self.filename}")
        temp_file = file_location + str(uuid4())
        if path is not None and not os.path.exists(path):
            raise DownloadFileError(None, f"No such directory exists: '{path}'")
        return file_location, temp_file

    def download(self, path: str = None) -> DownloadFileResponse:
        """

        Parameters
        ----------
        path : str
            Destination of the download file. Default is current working directory.

        Returns
        -------
        DownloadFileResponse

        """
        file_response = None
        http_response = None
        extension = ""
        file_location = ""
        temp_file = ""

        try:
            file_location, temp_file = self._prepare_and_validate_file_path(path)
            with open(temp_file, "wb") as f:
                with httpx.stream(method="GET", url=self.signed_url) as http_response:
                    if http_response.is_error:
                        # Raise Error
                        raise DownloadFileError(
                            http_response.status_code, http_response.read().decode()
                        )
                    # Process streaming data
                    for chunk in http_response.iter_bytes():
                        f.write(chunk)
                    # Check if streaming data is completely consumed
                    if http_response.is_stream_consumed:
                        file_response, extension = self._process_response(
                            file_location, http_response
                        )
                    else:
                        raise DownloadFileError(400, "File download is not completed.")
        except DownloadFileError as err:
            file_response = self._process_exception(http_response, err)
        except Exception as err:
            error = DownloadFileError(None, err.__str__())
            file_response = self._process_exception(http_response, error)

        if file_response is None:
            raise ValueError("File response is None!")

        if len(file_response.errors) > 0:
            if os.path.exists(temp_file):
                os.remove(temp_file)
            raise DownloadFileError(
                code=file_response.errors[0].code,
                message=file_response.errors[0].message,
            )
        else:
            file_location = file_location + extension
            if os.path.exists(file_location):
                os.remove(file_location)
            os.rename(temp_file, file_location)

        return file_response

    async def download_async(
        self, path: str = None, callback: Callable = None
    ) -> DownloadFileResponse:
        """

        Parameters
        ----------
        path : str
            Destination of the download file. Default is current working directory.
        callback : Callable
            Callback function will be called after the process is completed

        Returns
        -------
        DownloadFileResponse

        """
        file_response = None
        http_response = None
        extension = ""
        file_location = ""
        temp_file = ""
        client = httpx.AsyncClient()

        try:
            file_location, temp_file = self._prepare_and_validate_file_path(path)
            with open(temp_file, "wb") as f:
                async with client.stream(
                    method="GET", url=self.signed_url
                ) as http_response:
                    if http_response.is_error:
                        # Raise Error
                        error_message = await http_response.aread()
                        raise DownloadFileError(
                            http_response.status_code, error_message.decode()
                        )
                    # Process streaming data
                    async for chunk in http_response.aiter_bytes():
                        f.write(chunk)
                    # Check if streaming data is completely consumed
                    if http_response.is_stream_consumed:
                        file_response, extension = self._process_response(
                            file_location, http_response
                        )
                    else:
                        raise DownloadFileError(400, "File download is not completed.")
        except DownloadFileError as err:
            file_response = self._process_exception(http_response, err)
        except Exception as err:
            error = DownloadFileError(None, err.__str__())
            file_response = self._process_exception(http_response, error)

        await client.aclose()
        if file_response is None:
            return DownloadFileResponse(error=ValueError("File response is None!"))

        if len(file_response.errors) > 0:
            if os.path.exists(temp_file):
                os.remove(temp_file)
        else:
            file_location = file_location + extension
            if os.path.exists(file_location):
                os.remove(file_location)
            os.rename(temp_file, file_location)

        if callback is not None and callable(callback):
            callback(file_response)

        return file_response


class ListOfFile(list):
    # For download all file at once
    # response.data.files.download() while response.data.files is a list of file
    def download(self, path: str = None) -> DownloadAllFileResponse:
        """
        Parameters
        ----------
        path : str
            Destination of the download file. Default is current working directory.

        Returns
        -------
        DownloadAllFileResponse

        """
        # Download sync can return exception, so we have to use async download to get error response
        responses = asyncio.run(self.download_async(path=path))
        return responses

    async def download_async(
        self, path: str = None, callback: Callable = None
    ) -> DownloadAllFileResponse:
        """
        Parameters
        ----------
        path : str
            Destination of the download file. Default is current working directory.
        callback: Callable
            Callback function will be called after the process is completed

        Returns
        -------
        DownloadAllFileResponse

        """
        files = []
        errors = []

        if len(self) == 0:
            errors.append(
                DownloadFileError(
                    code=None, message="Cannot download any file. Files are empty."
                )
            )
            return DownloadAllFileResponse(files=files, errors=errors)

        responses = await asyncio.gather(
            *[each_file.download_async(path=path) for each_file in self]
        )
        for response in responses:
            files.extend(response.data.files)
            errors.extend(response.errors)
        errors = list(set(errors))
        download_all_response = DownloadAllFileResponse(files=files, errors=errors)

        if callback is not None and callable(callback):
            callback(download_all_response)

        return download_all_response


# ---------------------------------------------------------------------------
#   Response data
# ---------------------------------------------------------------------------


class FilingsData(Data):
    def __init__(self, raw, *args, **kwargs) -> None:
        super().__init__(raw=raw, *args, **kwargs)
        self._files = None

    @property
    def df(self):
        if self._dataframe is None and self._raw:
            # generate headers for df
            headers = [
                {"title": "Filename", "type": "string"},
                {"title": "SignedURL", "type": "string"},
                {"title": "MimeType", "type": "string"},
            ]

            # manipulate data for df
            """
            Example self._raw
            
            1. Request with filename
            {'signedUrl': 'https://cdn-filings.filings.refinitiv.com/retrieval/filings/ecpfilings_34359955599_pdf?ClientID=API_Playground&Expires=1648731287&Signature=asmxIEJpcNjF5DsbtGOVea-liyis3G53I2EJsbXJdr6wWt1~U~pntwcC7wbEvBWCLe7w0Oq0E6-NzeBkM8-W5ejIWYCsWOgGRekF1OPgBexZbVDULFqOqiFnk2tu6EXaZlExIU1DzYadnrxowDLbNSdlpYUmesWP9oTvOyyaBIFfvKaQ4dkjNYS9Bht2P7qSxqnQIGTELTAV~OIiM96JgY6eAYLdi0Yjv8TWZSqlDmapSZd~WFVnvrpKb24zt8g2Zg40XXKeI9lp1KuZ7qQqX1Bvi329Eg4L6Tx6lyBPivO1gf84xXHYyc-jSve4e5AF4FqKBDX0itWbODV1hFzdJQ__&Key-Pair-Id=APKAIDW27KNAZ6YUBN7A'}
            
            2. Request with dcn, doc_id, filing_id
            {
                'ecpfilings_97654291060_html': {
                    'signedUrl': 'https://cdn-filings.filings.refinitiv.com/retrieval/filings/ecpfilings_97654291060_html?ClientID=API_Playground&Expires=1648731287&Signature=Mj689Rgx4lMR4zMJ8ZBKW7vIVdYdl8Cy8bEkW84Lh1xe530GnzE~xHBuGV1AexrGP2QYa9lKO-DF5nVwnHzGm3RstsHtqsAFA5gbXFFp8qjgid-y-dDkV53z7YM0eePrb2fBtlKN5gSlY5jU-ueyIdiEKkhpSKQt05~SR3GXV5lPXBmyZJNEwjsVsmud6p5LjjWXPDFe1KexFSgjbtZ1QPvD5lyoPnGMJtGtPuUEtnBuJBel8Wj07LlLK9yLfJR-eMEJwreptO9IaOwFXZ78epP8~eYVSjcMFYhT5i3ku~I8OynVsBl9eA3vbWZ63-Z6qjotA93jf1hrIeYu8BdCmw__&Key-Pair-Id=APKAIDW27KNAZ6YUBN7A',
                    'mimeType': 'text/html'
                },
                'ecpfilings_97654291060_pdf': {
                    'signedUrl': 'https://cdn-filings.filings.refinitiv.com/retrieval/filings/ecpfilings_97654291060_pdf?ClientID=API_Playground&Expires=1648731287&Signature=dY~A~MSnyqTAErY2AAmHWNWOycncbGWnRvImAfxtXVd3Jz7DcofN-7rhZK6~5CEwtPY9oE1NLDlHhm4TbprAV7SEEG~h2n4iqZUtF7qC8bHQgiSTmwV10KIURLp-uto7aZIrsJQZHTmNJ4C1SrfhE041Ko6pxu-9hsMhF8VDC76mxy~l4eDqDzsxMOyGxNGMSl8a6sTNqlTGZyW61CnsUhOLGiGVfNz3itKv54GWsudAhNOtS1DwcDr~D-5b3dinJduyH4XmOJgKvMJlPxxI5ciBKZ-k2GF~2~bl08gm-XVxLyrlYdtlHPB6KohZDEgppvs921yZuOZ--NMBMPl4zg__&Key-Pair-Id=APKAIDW27KNAZ6YUBN7A',
                    'mimeType': 'application/pdf'
                },
                'ecpfilings_97654291060_dissemination_txt': {
                    'signedUrl': 'https://cdn-filings.filings.refinitiv.com/retrieval/filings/ecpfilings_97654291060_dissemination_txt?ClientID=API_Playground&Expires=1648731287&Signature=e-FxtjqWnh~Q20uL4xaNXxlZhlP7Gvqqa1uVHsbIOFgpVqZd39oKUnyNn1do8MdpUYveP~YRvHzJ8uAYncjY1NaMHixYmfoZxqcfAHasmdMwdtaAzihPxmZobLHu7kK0iIfpCglF9RH2e7yT-WhglXbdUDvlY~eVnAFPMf4Q-tkGPQGAtmR1pZvYX53GCo-XIHO3-bX-YcY4MbsAQdzqKNdaHVKySZ0RoyfJsjdOKmJFmEygfoKQvg3zL-HV2FsD9uCZXPlV9elN3OnGyXoOSOmX4unroh7vwI5NBV3pO5x37JIl~WE4a9KU6~sacGpXiz8Sg~RkaiyQMbSUR1tsVA__&Key-Pair-Id=APKAIDW27KNAZ6YUBN7A',
                    'mimeType': 'text/plain'
                }
            }
            """
            if "signedUrl" in self._raw:
                # Need to extract filename from signedURL
                url_parse = urllib.parse.urlparse(self._raw["signedUrl"])

                # url_parse.path can be '/retrieval/filings/ecpfilings_34359955599_pdf'
                # [[Filename, SignedURL, MimeType]]
                # [["ecpfilings_34359955599_pdf", self._raw["signedUrl"], ""]]
                data = [[url_parse.path.split("/")[-1], self._raw["signedUrl"], ""]]
            else:
                # we can get filename from key in self._raw in this case
                data = [
                    [filename, attributes["signedUrl"], attributes["mimeType"]]
                    for filename, attributes in self._raw.items()
                ]

            self._raw.update({"headers": headers, "data": data})
            self._dataframe = self._dfbuilder(self._raw, **self._kwargs)
        return self._dataframe

    @property
    def files(self):
        if self._files is None:
            self._files = ListOfFile()
            if self.df is not None and not self.df.empty:
                self._files.extend(FilingsFile(*row) for row in self.df.values)
        return self._files


class FilingsResponse(BaseResponse[FilingsData]):
    pass


# ---------------------------------------------------------------------------
#   Provider layer
# ---------------------------------------------------------------------------


class FilingsDataProviderLayer(ContentProviderLayer):
    def _check_response(self, response, config):
        _check_response(response, config, response_class=FilingsResponse)

    def get_data(self, session=None, on_response=None, **kwargs):
        """
        Returns a response to the data platform

        Parameters
        ----------
        session : Session, optional
            Means default session would be used
        on_response : Callable, optional
            Callable object to process retrieved data

        Returns
        -------
        FilingsResponse

        Raises
        ------
        AttributeError
            If user didn't set default session.

        """
        return super().get_data(session, on_response, **kwargs)

    async def get_data_async(self, session=None, on_response=None, **kwargs):
        """
        Returns a response to the data platform

        Parameters
        ----------
        session : Session, optional
            Means default session would be used
        on_response : Callable, optional
            Callable object to process retrieved data

        Returns
        -------
        FilingsResponse

        Raises
        ------
        AttributeError
            If user didn't set default session.

        """
        return await super().get_data_async(session, on_response, **kwargs)


filings_data_provider = DataProvider(
    request=FilingsRequestFactory(),
    response=ResponseFactory(response_class=FilingsResponse, data_class=FilingsData),
    validator=UniverseContentValidator(),
    parser=ErrorParser(),
)



# coding: utf8

__all__ = ["CreditSpreadType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class CreditSpreadType(Enum):
    FLAT_SPREAD = "FlatSpread"
    TERM_STRUCTURE = "TermStructure"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(CreditSpreadType, _CREDIT_SPREAD_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(
            _CREDIT_SPREAD_TYPE_VALUES_IN_LOWER_BY_CREDIT_SPREAD_TYPE, some
        )


_CREDIT_SPREAD_TYPE_VALUES = tuple(t.value for t in CreditSpreadType)
_CREDIT_SPREAD_TYPE_VALUES_IN_LOWER_BY_CREDIT_SPREAD_TYPE = {
    name.lower(): item for name, item in CreditSpreadType.__members__.items()
}



import itertools
from typing import (
    Union,
    Callable,
    Iterable,
    Tuple,
    TYPE_CHECKING,
)

import pandas as pd

from .._content_type import ContentType
from ..._core.session import get_valid_session, Session
from ..._tools import DEBUG, cached_property, PRICING_DATETIME_PATTERN
from ..._tools._dataframe import convert_df_columns_to_datetime_use_re_compile
from ...delivery._stream import (
    StreamStateManager,
    OMMStreamListener,
    StreamStateEvent,
    _OMMStream,
    StreamState,
)
from ...delivery._stream._stream_factory import create_omm_stream
from ...delivery._stream.base_stream import stream_state_to_open_state
from ...delivery._stream.stream_cache import StreamCache

if TYPE_CHECKING:
    from ... import OpenState

_id_iterator = itertools.count()


class CustomInstrumentsStream(
    StreamCache, StreamStateManager, OMMStreamListener["CustomInstrumentsStream"]
):
    def __init__(
        self,
        name: Union[str, Iterable[str]],
        session: "Session" = None,
        service=None,
        on_refresh: Callable = None,
        on_status: Callable = None,
        on_update: Callable = None,
        on_complete: Callable = None,
        extended_params: dict = None,
    ):
        if name is None:
            raise AttributeError("universe name must be defined.")

        session = get_valid_session(session)

        StreamCache.__init__(self, name=name, service=service)
        StreamStateManager.__init__(self, logger=session.logger())
        OMMStreamListener.__init__(
            self,
            logger=session.logger(),
            on_refresh=on_refresh,
            on_status=on_status,
            on_update=on_update,
            on_complete=on_complete,
        )

        self._session = session
        self._name = name
        self._extended_params = extended_params
        self._record = {}

    @cached_property
    def _stream(self):
        stream = create_omm_stream(
            ContentType.STREAMING_CUSTOM_INSTRUMENTS,
            session=self._session,
            name=self._name,
            domain="MarketPrice",
            extended_params=self._extended_params,
            on_refresh=self._on_stream_refresh,
            on_status=self._on_stream_status,
            on_update=self._on_stream_update,
            on_complete=self._on_stream_complete,
            on_error=self._on_stream_error,
        )
        stream.on(StreamStateEvent.CLOSED, self.close)
        return stream

    def _do_on_stream_update(self, stream: _OMMStream, *args):
        message = args[0]

        if DEBUG:
            fields = self._record.get("Fields", [])
            num_fields = len(fields)
            self._debug(
                f"|>|>|>|>|>|> {self._classname} "
                f"has fields in record {num_fields} after update"
            )

        fields = message.get("Fields")
        self._record["Fields"].update(fields)
        return fields

    def _do_open(self, *args, with_updates=True):
        self._stream.open(*args, with_updates=with_updates)

    def _do_pause(self):
        self._stream.pause()

    def _do_resume(self):
        self._stream.resume()

    def _do_close(self, *args, **kwargs):
        self._stream.close(*args, **kwargs)

    def _do_on_stream_refresh(self, stream: "_OMMStream", *args) -> Tuple:
        message = args[0]
        self._record = message

        if DEBUG:
            fields = self._record.get("Fields", [])
            num_fields = len(fields)
            self._debug(
                f"|>|>|>|>|>|>{self._classname} "
                f"has fields in record {num_fields} after refresh"
            )

        fields = message.get("Fields", {})
        return fields

    def get_snapshot(self):
        fields = self._record.get("Fields")
        df = pd.DataFrame(fields, index=[self._name])
        convert_df_columns_to_datetime_use_re_compile(df, PRICING_DATETIME_PATTERN)
        return df

    @property
    def open_state(self) -> "OpenState":
        state: "StreamState" = self._stream.state
        state: "OpenState" = stream_state_to_open_state.get(state)
        return state



# coding: utf8

from enum import Enum, unique

from ..._enums._common_tools import _convert_to_str, _normalize


@unique
class SurfaceOutputs(Enum):
    HEADERS = "Headers"
    DATATYPE = "DataType"
    DATA = "Data"
    STATUSES = "Statuses"
    FORWARD_CURVE = "ForwardCurve"
    DIVIDENDS = "Dividends"
    INTEREST_RATE_CURVE = "InterestRateCurve"
    GOODNESS_OF_FIT = "GoodnessOfFit"
    UNDERLYING_SPOT = "UnderlyingSpot"
    DISCOUNT_CURVE = "DiscountCurve"
    MONEYNESS_STRIKE = "MoneynessStrike"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(SurfaceOutputs, _SURFACE_OUTPUTS_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_SURFACE_OUTPUTS_VALUES_IN_LOWER_BY_SURFACE_OUTPUTS, some)


_SURFACE_OUTPUTS_VALUES = tuple(t.value for t in SurfaceOutputs)
_SURFACE_OUTPUTS_VALUES_IN_LOWER_BY_SURFACE_OUTPUTS = {
    name.lower(): item for name, item in SurfaceOutputs.__members__.items()
}



# coding: utf8
# contract_gen 2020-05-18 08:30:59.240817

__all__ = ["LocalVolatilityMethod"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class LocalVolatilityMethod(Enum):
    BEST_SMILE = "BestSmile"
    CONVEX_SMILE = "ConvexSmile"
    PARABOLA_SMOOTH = "ParabolaSmooth"
    PARABOLA_WITHOUT_EXTRAPOL = "ParabolaWithoutExtrapol"
    RATIONAL_SMOOTH = "RationalSmooth"
    RATIONAL_WITHOUT_EXTRAPOL = "RationalWithoutExtrapol"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            LocalVolatilityMethod, _LOCAL_VOLATILITY_METHOD_VALUES, some
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _LOCAL_VOLATILITY_METHOD_VALUES_IN_LOWER_BY_LOCAL_VOLATILITY_METHOD, some
        )


_LOCAL_VOLATILITY_METHOD_VALUES = tuple(t.value for t in LocalVolatilityMethod)
_LOCAL_VOLATILITY_METHOD_VALUES_IN_LOWER_BY_LOCAL_VOLATILITY_METHOD = {
    name.lower(): item for name, item in LocalVolatilityMethod.__members__.items()
}
